(-f0.0) iff(16) Jump: 5                                         { align1 switch };
(+f0.0.x) iff(8) Jump: 11                                       { align16 switch };
(+f0.0) iff(16) Jump: 7                                         { align1 switch };
