(-f0.1.any4h) halt(8) JIP: LABEL0      UIP: LABEL0              { align1 1Q };
halt(8)         JIP: LABEL1            UIP: LABEL1              { align1 1Q };
LABEL1:
(-f0.1.any4h) halt(16) JIP: LABEL0     UIP: LABEL0              { align1 1H };
halt(16)        JIP: LABEL0            UIP: LABEL0              { align1 1H };
LABEL0: