LABEL0:
while(8)        JIP: LABEL0                                     { align1 1Q };
while(16)       JIP: LABEL0                                     { align1 1H };
while(8)        JIP: LABEL0                                     { align16 1Q };
(-f0.0) while(8) JIP: LABEL0                                    { align1 1Q };
(-f0.0) while(16) JIP: LABEL0                                   { align1 1H };
