(+f0.0) if(8)   JIP: LABEL0       UIP: LABEL1                   { align1 1Q };
(-f0.0) if(8)   JIP: LABEL0       UIP: LABEL1                   { align1 1Q };
LABEL0:
(-f0.0) if(16)  JIP: LABEL1       UIP: LABEL1                   { align1 1H };
(+f0.0) if(16)  JIP: LABEL1       UIP: LABEL1                   { align1 1H };
(+f0.0) if(32)  JIP: LABEL1       UIP: LABEL1                   { align1 };
LABEL1:
