/*
 * Copyright © 2022 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include "anv_private.h"

VkResult
hitman3_CreateBufferView(VkDevice _device,
                         const VkBufferViewCreateInfo *pCreateInfo,
                         const VkAllocationCallbacks *pAllocator,
                         VkBufferView *pView)
{
   ANV_FROM_HANDLE(anv_buffer, buffer, pCreateInfo->buffer);
   if (pCreateInfo->format == VK_FORMAT_R32G32B32_SFLOAT &&
       (buffer->vk.usage & VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT)) {
      ANV_FROM_HANDLE(anv_device, device, _device);
      return vk_errorf(device, VK_ERROR_UNKNOWN,
                       "invalid image format requested for storage");
   }

   return anv_CreateBufferView(_device, pCreateInfo, pAllocator, pView);
}
