#ifndef GM107_TEXTURE_XML
#define GM107_TEXTURE_XML

/* Autogenerated file, DO NOT EDIT manually!

This file was generated by the rules-ng-ng headergen tool in this git repository:
http://github.com/envytools/envytools/
git clone https://github.com/envytools/envytools.git

The rules-ng-ng source files this header was generated from are:
- /home/skeggsb/git/envytools/rnndb/../rnndb/graph/gm107_texture.xml (  22057 bytes, from 2016-02-12 03:01:43)
- /home/skeggsb/git/envytools/rnndb/copyright.xml                    (   6456 bytes, from 2015-09-10 02:57:40)
- /home/skeggsb/git/envytools/rnndb/nvchipsets.xml                   (   2908 bytes, from 2016-02-04 22:19:11)
- /home/skeggsb/git/envytools/rnndb/g80_defs.xml                     (  21739 bytes, from 2016-02-04 00:29:42)

Copyright (C) 2006-2016 by the following authors:
- Artur Huillet <arthur.huillet@free.fr> (ahuillet)
- Ben Skeggs (darktama, darktama_)
- B. R. <koala_br@users.sourceforge.net> (koala_br)
- Carlos Martin <carlosmn@users.sf.net> (carlosmn)
- Christoph Bumiller <e0425955@student.tuwien.ac.at> (calim, chrisbmr)
- Dawid Gajownik <gajownik@users.sf.net> (gajownik)
- Dmitry Baryshkov
- Dmitry Eremin-Solenikov <lumag@users.sf.net> (lumag)
- EdB <edb_@users.sf.net> (edb_)
- Erik Waling <erikwailing@users.sf.net> (erikwaling)
- Francisco Jerez <currojerez@riseup.net> (curro)
- Ilia Mirkin <imirkin@alum.mit.edu> (imirkin)
- jb17bsome <jb17bsome@bellsouth.net> (jb17bsome)
- Jeremy Kolb <kjeremy@users.sf.net> (kjeremy)
- Laurent Carlier <lordheavym@gmail.com> (lordheavy)
- Luca Barbieri <luca@luca-barbieri.com> (lb, lb1)
- Maarten Maathuis <madman2003@gmail.com> (stillunknown)
- Marcin Kościelnicki <koriakin@0x04.net> (mwk, koriakin)
- Mark Carey <mark.carey@gmail.com> (careym)
- Matthieu Castet <matthieu.castet@parrot.com> (mat-c)
- nvidiaman <nvidiaman@users.sf.net> (nvidiaman)
- Patrice Mandin <patmandin@gmail.com> (pmandin, pmdata)
- Pekka Paalanen <pq@iki.fi> (pq, ppaalanen)
- Peter Popov <ironpeter@users.sf.net> (ironpeter)
- Richard Hughes <hughsient@users.sf.net> (hughsient)
- Rudi Cilibrasi <cilibrar@users.sf.net> (cilibrar)
- Serge Martin
- Simon Raffeiner
- Stephane Loeuillet <leroutier@users.sf.net> (leroutier)
- Stephane Marchesin <stephane.marchesin@gmail.com> (marcheu)
- sturmflut <sturmflut@users.sf.net> (sturmflut)
- Sylvain Munaut <tnt@246tNt.com>
- Victor Stinner <victor.stinner@haypocalc.com> (haypo)
- Wladmir van der Laan <laanwj@gmail.com> (miathan6)
- Younes Manton <younes.m@gmail.com> (ymanton)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/


#define GM107_TIC2__SIZE					0x00000020
#define GM107_TIC2_0						0x00000000
#define GM107_TIC2_0_COMPONENTS_SIZES__MASK			0x0000007f
#define GM107_TIC2_0_COMPONENTS_SIZES__SHIFT			0
#define GM107_TIC2_0_COMPONENTS_SIZES_R32_G32_B32_A32		0x00000001
#define GM107_TIC2_0_COMPONENTS_SIZES_R32_G32_B32		0x00000002
#define GM107_TIC2_0_COMPONENTS_SIZES_R16_G16_B16_A16		0x00000003
#define GM107_TIC2_0_COMPONENTS_SIZES_R32_G32			0x00000004
#define GM107_TIC2_0_COMPONENTS_SIZES_R32_B24G8			0x00000005
#define GM107_TIC2_0_COMPONENTS_SIZES_X8B8G8R8			0x00000007
#define GM107_TIC2_0_COMPONENTS_SIZES_A8B8G8R8			0x00000008
#define GM107_TIC2_0_COMPONENTS_SIZES_A2B10G10R10		0x00000009
#define GM107_TIC2_0_COMPONENTS_SIZES_R16_G16			0x0000000c
#define GM107_TIC2_0_COMPONENTS_SIZES_G8R24			0x0000000d
#define GM107_TIC2_0_COMPONENTS_SIZES_G24R8			0x0000000e
#define GM107_TIC2_0_COMPONENTS_SIZES_R32			0x0000000f
#define GM107_TIC2_0_COMPONENTS_SIZES_A4B4G4R4			0x00000012
#define GM107_TIC2_0_COMPONENTS_SIZES_A5B5G5R1			0x00000013
#define GM107_TIC2_0_COMPONENTS_SIZES_A1B5G5R5			0x00000014
#define GM107_TIC2_0_COMPONENTS_SIZES_B5G6R5			0x00000015
#define GM107_TIC2_0_COMPONENTS_SIZES_B6G5R5			0x00000016
#define GM107_TIC2_0_COMPONENTS_SIZES_G8R8			0x00000018
#define GM107_TIC2_0_COMPONENTS_SIZES_R16			0x0000001b
#define GM107_TIC2_0_COMPONENTS_SIZES_Y8_VIDEO			0x0000001c
#define GM107_TIC2_0_COMPONENTS_SIZES_R8			0x0000001d
#define GM107_TIC2_0_COMPONENTS_SIZES_G4R4			0x0000001e
#define GM107_TIC2_0_COMPONENTS_SIZES_R1			0x0000001f
#define GM107_TIC2_0_COMPONENTS_SIZES_E5B9G9R9_SHAREDEXP	0x00000020
#define GM107_TIC2_0_COMPONENTS_SIZES_BF10GF11RF11		0x00000021
#define GM107_TIC2_0_COMPONENTS_SIZES_G8B8G8R8			0x00000022
#define GM107_TIC2_0_COMPONENTS_SIZES_B8G8R8G8			0x00000023
#define GM107_TIC2_0_COMPONENTS_SIZES_DXT1			0x00000024
#define GM107_TIC2_0_COMPONENTS_SIZES_DXT23			0x00000025
#define GM107_TIC2_0_COMPONENTS_SIZES_DXT45			0x00000026
#define GM107_TIC2_0_COMPONENTS_SIZES_DXN1			0x00000027
#define GM107_TIC2_0_COMPONENTS_SIZES_DXN2			0x00000028
#define GM107_TIC2_0_COMPONENTS_SIZES_BC6H_SF16			0x00000010
#define GM107_TIC2_0_COMPONENTS_SIZES_BC6H_UF16			0x00000011
#define GM107_TIC2_0_COMPONENTS_SIZES_BC7U			0x00000017
#define GM107_TIC2_0_COMPONENTS_SIZES_ETC2_RGB			0x00000006
#define GM107_TIC2_0_COMPONENTS_SIZES_ETC2_RGB_PTA		0x0000000a
#define GM107_TIC2_0_COMPONENTS_SIZES_ETC2_RGBA			0x0000000b
#define GM107_TIC2_0_COMPONENTS_SIZES_EAC			0x00000019
#define GM107_TIC2_0_COMPONENTS_SIZES_EACX2			0x0000001a
#define GM107_TIC2_0_COMPONENTS_SIZES_Z24S8			0x00000029
#define GM107_TIC2_0_COMPONENTS_SIZES_X8Z24			0x0000002a
#define GM107_TIC2_0_COMPONENTS_SIZES_S8Z24			0x0000002b
#define GM107_TIC2_0_COMPONENTS_SIZES_X4V4Z24__COV4R4V		0x0000002c
#define GM107_TIC2_0_COMPONENTS_SIZES_X4V4Z24__COV8R8V		0x0000002d
#define GM107_TIC2_0_COMPONENTS_SIZES_V8Z24__COV4R12V		0x0000002e
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32			0x0000002f
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X24S8		0x00000030
#define GM107_TIC2_0_COMPONENTS_SIZES_X8Z24_X20V4S8__COV4R4V	0x00000031
#define GM107_TIC2_0_COMPONENTS_SIZES_X8Z24_X20V4S8__COV8R8V	0x00000032
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X20V4X8__COV4R4V	0x00000033
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X20V4X8__COV8R8V	0x00000034
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X20V4S8__COV4R4V	0x00000035
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X20V4S8__COV8R8V	0x00000036
#define GM107_TIC2_0_COMPONENTS_SIZES_X8Z24_X16V8S8__COV4R12V	0x00000037
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X16V8X8__COV4R12V	0x00000038
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X16V8S8__COV4R12V	0x00000039
#define GM107_TIC2_0_COMPONENTS_SIZES_Z16			0x0000003a
#define GM107_TIC2_0_COMPONENTS_SIZES_V8Z24__COV8R24V		0x0000003b
#define GM107_TIC2_0_COMPONENTS_SIZES_X8Z24_X16V8S8__COV8R24V	0x0000003c
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X16V8X8__COV8R24V	0x0000003d
#define GM107_TIC2_0_COMPONENTS_SIZES_ZF32_X16V8S8__COV8R24V	0x0000003e
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_4X4		0x00000040
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_5X4		0x00000050
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_5X5		0x00000041
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_6X5		0x00000051
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_6X6		0x00000042
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_8X5		0x00000055
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_8X6		0x00000052
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_8X8		0x00000044
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_10X5		0x00000056
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_10X6		0x00000057
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_10X8		0x00000053
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_10X10		0x00000045
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_12X10		0x00000054
#define GM107_TIC2_0_COMPONENTS_SIZES_ASTC_2D_12X12		0x00000046
#define GM107_TIC2_0_R_DATA_TYPE__MASK				0x00000380
#define GM107_TIC2_0_R_DATA_TYPE__SHIFT				7
#define GM107_TIC2_0_G_DATA_TYPE__MASK				0x00001c00
#define GM107_TIC2_0_G_DATA_TYPE__SHIFT				10
#define GM107_TIC2_0_B_DATA_TYPE__MASK				0x0000e000
#define GM107_TIC2_0_B_DATA_TYPE__SHIFT				13
#define GM107_TIC2_0_A_DATA_TYPE__MASK				0x00070000
#define GM107_TIC2_0_A_DATA_TYPE__SHIFT				16
#define GM107_TIC2_0_X_SOURCE__MASK				0x00380000
#define GM107_TIC2_0_X_SOURCE__SHIFT				19
#define GM107_TIC2_0_Y_SOURCE__MASK				0x01c00000
#define GM107_TIC2_0_Y_SOURCE__SHIFT				22
#define GM107_TIC2_0_Z_SOURCE__MASK				0x0e000000
#define GM107_TIC2_0_Z_SOURCE__SHIFT				25
#define GM107_TIC2_0_W_SOURCE__MASK				0x70000000
#define GM107_TIC2_0_W_SOURCE__SHIFT				28
#define GM107_TIC2_0_PACK_COMPONENTS				0x80000000

#define GM107_TIC2_1						0x00000004
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_0__MASK			0xffffffff
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_0__SHIFT		0
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_5__MASK			0xffffffe0
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_5__SHIFT		5
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_5__SHR			5
#define GM107_TIC2_1_GOB_DEPTH_OFFSET__MASK			0x00000060
#define GM107_TIC2_1_GOB_DEPTH_OFFSET__SHIFT			5
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_9__MASK			0xfffffe00
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_9__SHIFT		9
#define GM107_TIC2_1_ADDRESS_BITS_31_TO_9__SHR			9

#define GM107_TIC2_2						0x00000008
#define GM107_TIC2_2_ADDRESS_BITS_47_TO_32__MASK		0x0000ffff
#define GM107_TIC2_2_ADDRESS_BITS_47_TO_32__SHIFT		0
#define GM107_TIC2_2_HEADER_VERSION__MASK			0x00e00000
#define GM107_TIC2_2_HEADER_VERSION__SHIFT			21
#define GM107_TIC2_2_HEADER_VERSION_ONE_D_BUFFER		0x00000000
#define GM107_TIC2_2_HEADER_VERSION_PITCH_COLORKEY		0x00200000
#define GM107_TIC2_2_HEADER_VERSION_PITCH			0x00400000
#define GM107_TIC2_2_HEADER_VERSION_BLOCKLINEAR			0x00600000
#define GM107_TIC2_2_HEADER_VERSION_BLOCKLINEAR_COLORKEY	0x00800000
#define GM107_TIC2_2_RESOURCE_VIEW_COHERENCY_HASH__MASK		0x1e000000
#define GM107_TIC2_2_RESOURCE_VIEW_COHERENCY_HASH__SHIFT	25

#define GM107_TIC2_3						0x0000000c
#define GM107_TIC2_3_WIDTH_MINUS_ONE_BITS_31_TO_16__MASK	0x0000ffff
#define GM107_TIC2_3_WIDTH_MINUS_ONE_BITS_31_TO_16__SHIFT	0
#define GM107_TIC2_3_PITCH_BITS_20_TO_5__MASK			0x0000ffff
#define GM107_TIC2_3_PITCH_BITS_20_TO_5__SHIFT			0
#define GM107_TIC2_3_PITCH_BITS_20_TO_5__SHR			5
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH__MASK			0x00000007
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH__SHIFT		0
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH__MIN			0x00000000
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH__MAX			0x00000000
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_ONE			0x00000000
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_TWO			0x00000001
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_FOUR			0x00000002
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_EIGHT			0x00000003
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_SIXTEEN		0x00000004
#define GM107_TIC2_3_GOBS_PER_BLOCK_WIDTH_THIRTYTWO		0x00000005
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT__MASK		0x00000038
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT__SHIFT		3
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_ONE			0x00000000
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_TWO			0x00000008
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_FOUR			0x00000010
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_EIGHT		0x00000018
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_SIXTEEN		0x00000020
#define GM107_TIC2_3_GOBS_PER_BLOCK_HEIGHT_THIRTYTWO		0x00000028
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH__MASK			0x000001c0
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH__SHIFT		6
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_ONE			0x00000000
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_TWO			0x00000040
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_FOUR			0x00000080
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_EIGHT			0x000000c0
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_SIXTEEN		0x00000100
#define GM107_TIC2_3_GOBS_PER_BLOCK_DEPTH_THIRTYTWO		0x00000140
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS__MASK			0x00001c00
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS__SHIFT			10
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_ONE			0x00000000
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_TWO			0x00000400
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_FOUR			0x00000800
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_EIGHT			0x00000c00
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_SIXTEEN			0x00001000
#define GM107_TIC2_3_TILE_WIDTH_IN_GOBS_THIRTYTWO		0x00001400
#define GM107_TIC2_3_GOB_3D					0x00002000
#define GM107_TIC2_3_LOD_ANISO_QUALITY_2			0x00010000
#define GM107_TIC2_3_LOD_ANISO_QUALITY__MASK			0x00020000
#define GM107_TIC2_3_LOD_ANISO_QUALITY__SHIFT			17
#define GM107_TIC2_3_LOD_ANISO_QUALITY_LOW			0x00000000
#define GM107_TIC2_3_LOD_ANISO_QUALITY_HIGH			0x00020000
#define GM107_TIC2_3_LOD_ISO_QUALITY__MASK			0x00040000
#define GM107_TIC2_3_LOD_ISO_QUALITY__SHIFT			18
#define GM107_TIC2_3_LOD_ISO_QUALITY_LOW			0x00000000
#define GM107_TIC2_3_LOD_ISO_QUALITY_HIGH			0x00040000
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER__MASK		0x00180000
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER__SHIFT	19
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER_NONE		0x00000000
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER_CONST_ONE	0x00080000
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER_CONST_TWO	0x00100000
#define GM107_TIC2_3_ANISO_COARSE_SPREAD_MODIFIER_SQRT		0x00180000
#define GM107_TIC2_3_ANISO_SPREAD_SCALE__MASK			0x03e00000
#define GM107_TIC2_3_ANISO_SPREAD_SCALE__SHIFT			21
#define GM107_TIC2_3_USE_HEADER_OPT_CONTROL			0x04000000
#define GM107_TIC2_3_DEPTH_TEXTURE				0x08000000
#define GM107_TIC2_3_MAX_MIP_LEVEL__MASK			0xf0000000
#define GM107_TIC2_3_MAX_MIP_LEVEL__SHIFT			28

#define GM107_TIC2_4						0x00000010
#define GM107_TIC2_4_WIDTH_MINUS_ONE_BITS_15_TO_0__MASK		0x0000ffff
#define GM107_TIC2_4_WIDTH_MINUS_ONE_BITS_15_TO_0__SHIFT	0
#define GM107_TIC2_4_WIDTH_MINUS_ONE__MASK			0x0000ffff
#define GM107_TIC2_4_WIDTH_MINUS_ONE__SHIFT			0
#define GM107_TIC2_4_ANISO_SPREAD_MAX_LOG2__MASK		0x00380000
#define GM107_TIC2_4_ANISO_SPREAD_MAX_LOG2__SHIFT		19
#define GM107_TIC2_4_SRGB_CONVERSION				0x00400000
#define GM107_TIC2_4_TEXTURE_TYPE__MASK				0x07800000
#define GM107_TIC2_4_TEXTURE_TYPE__SHIFT			23
#define GM107_TIC2_4_TEXTURE_TYPE_ONE_D				0x00000000
#define GM107_TIC2_4_TEXTURE_TYPE_TWO_D				0x00800000
#define GM107_TIC2_4_TEXTURE_TYPE_THREE_D			0x01000000
#define GM107_TIC2_4_TEXTURE_TYPE_CUBEMAP			0x01800000
#define GM107_TIC2_4_TEXTURE_TYPE_ONE_D_ARRAY			0x02000000
#define GM107_TIC2_4_TEXTURE_TYPE_TWO_D_ARRAY			0x02800000
#define GM107_TIC2_4_TEXTURE_TYPE_ONE_D_BUFFER			0x03000000
#define GM107_TIC2_4_TEXTURE_TYPE_TWO_D_NO_MIPMAP		0x03800000
#define GM107_TIC2_4_TEXTURE_TYPE_CUBE_ARRAY			0x04000000
#define GM107_TIC2_4_SECTOR_PROMOTION__MASK			0x18000000
#define GM107_TIC2_4_SECTOR_PROMOTION__SHIFT			27
#define GM107_TIC2_4_SECTOR_PROMOTION_NO_PROMOTION		0x00000000
#define GM107_TIC2_4_SECTOR_PROMOTION_PROMOTE_TO_2_V		0x08000000
#define GM107_TIC2_4_SECTOR_PROMOTION_PROMOTE_TO_2_H		0x10000000
#define GM107_TIC2_4_SECTOR_PROMOTION_PROMOTE_TO_4		0x18000000
#define GM107_TIC2_4_BORDER_SIZE__MASK				0xe0000000
#define GM107_TIC2_4_BORDER_SIZE__SHIFT				29
#define GM107_TIC2_4_BORDER_SIZE_ONE				0x00000000
#define GM107_TIC2_4_BORDER_SIZE_TWO				0x20000000
#define GM107_TIC2_4_BORDER_SIZE_FOUR				0x40000000
#define GM107_TIC2_4_BORDER_SIZE_EIGHT				0x60000000
#define GM107_TIC2_4_BORDER_SIZE_SAMPLER_COLOR			0xe0000000

#define GM107_TIC2_5						0x00000014
#define GM107_TIC2_5_HEIGHT_MINUS_ONE__MASK			0x0000ffff
#define GM107_TIC2_5_HEIGHT_MINUS_ONE__SHIFT			0
#define GM107_TIC2_5_DEPTH_MINUS_ONE__MASK			0x3fff0000
#define GM107_TIC2_5_DEPTH_MINUS_ONE__SHIFT			16
#define GM107_TIC2_5_NORMALIZED_COORDS				0x80000000

#define GM107_TIC2_6						0x00000018
#define GM107_TIC2_6_COLOR_KEY_OP				0x00000001
#define GM107_TIC2_6_TRILIN_OPT__MASK				0x0000003e
#define GM107_TIC2_6_TRILIN_OPT__SHIFT				1
#define GM107_TIC2_6_MIP_LOD_BIAS__MASK				0x0007ffc0
#define GM107_TIC2_6_MIP_LOD_BIAS__SHIFT			6
#define GM107_TIC2_6_MIP_LOD_BIAS__RADIX			0x00000008
#define GM107_TIC2_6_ANISO_BIAS__MASK				0x00780000
#define GM107_TIC2_6_ANISO_BIAS__SHIFT				19
#define GM107_TIC2_6_ANISO_BIAS__RADIX				0x00000004
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC__MASK		0x01800000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC__SHIFT		23
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC_HALF		0x00000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC_ONE			0x00800000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC_TWO			0x01000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_FUNC_MAX			0x01800000
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC__MASK		0x06000000
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC__SHIFT		25
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC_HALF		0x00000000
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC_ONE		0x02000000
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC_TWO		0x04000000
#define GM107_TIC2_6_ANISO_COARSE_SPREAD_FUNC_MAX		0x06000000
#define GM107_TIC2_6_MAX_ANISOTROPY__MASK			0x38000000
#define GM107_TIC2_6_MAX_ANISOTROPY__SHIFT			27
#define GM107_TIC2_6_MAX_ANISOTROPY_1_TO_1			0x00000000
#define GM107_TIC2_6_MAX_ANISOTROPY_2_TO_1			0x08000000
#define GM107_TIC2_6_MAX_ANISOTROPY_4_TO_1			0x10000000
#define GM107_TIC2_6_MAX_ANISOTROPY_6_TO_1			0x18000000
#define GM107_TIC2_6_MAX_ANISOTROPY_8_TO_1			0x20000000
#define GM107_TIC2_6_MAX_ANISOTROPY_10_TO_1			0x28000000
#define GM107_TIC2_6_MAX_ANISOTROPY_12_TO_1			0x30000000
#define GM107_TIC2_6_MAX_ANISOTROPY_16_TO_1			0x38000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER__MASK		0xc0000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER__SHIFT		30
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER_NONE		0x00000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER_CONST_ONE	0x40000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER_CONST_TWO	0x80000000
#define GM107_TIC2_6_ANISO_FINE_SPREAD_MODIFIER_SQRT		0xc0000000

#define GM107_TIC2_7						0x0000001c
#define GM107_TIC2_7_COLOR_KEY_VALUE__MASK			0xffffffff
#define GM107_TIC2_7_COLOR_KEY_VALUE__SHIFT			0
#define GM107_TIC2_7_RES_VIEW_MIN_MIP_LEVEL__MASK		0x0000000f
#define GM107_TIC2_7_RES_VIEW_MIN_MIP_LEVEL__SHIFT		0
#define GM107_TIC2_7_RES_VIEW_MAX_MIP_LEVEL__MASK		0x000000f0
#define GM107_TIC2_7_RES_VIEW_MAX_MIP_LEVEL__SHIFT		4
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT__MASK			0x00000f00
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT__SHIFT			8
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_1X1			0x00000000
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_2X1			0x00000100
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_2X2			0x00000200
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_4X2			0x00000300
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_4X2_D3D			0x00000400
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_2X1_D3D			0x00000500
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_4X4			0x00000600
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_2X2_VC_4		0x00000800
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_2X2_VC_12		0x00000900
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_4X2_VC_8		0x00000a00
#define GM107_TIC2_7_MULTI_SAMPLE_COUNT_4X2_VC_24		0x00000b00
#define GM107_TIC2_7_MIN_LOD_CLAMP__MASK			0x00fff000
#define GM107_TIC2_7_MIN_LOD_CLAMP__SHIFT			12
#define GM107_TIC2_7_MIN_LOD_CLAMP__RADIX			0x00000008


#endif /* GM107_TEXTURE_XML */
