(+f0.0) if(8)   Jump: 10                                        { align1 switch };
(+f0.0) if(16)  Jump: 10                                        { align1 switch };
(+f0.0.x) if(8) Jump: 26                                        { align16 switch };
