CHIPSET(0x6780, TAHITI)
CHIPSET(0x6784, TAHITI)
CHIPSET(0x6788, TAHITI)
CHIPSET(0x678A, TAHITI)
CHIPSET(0x6790, TAHITI)
CHIPSET(0x6791, TAHITI)
CHIPSET(0x6792, TAHITI)
CHIPSET(0x6798, TAHITI)
CHIPSET(0x6799, TAHITI)
CHIPSET(0x679A, TAHITI)
CHIPSET(0x679B, TAHITI)
CHIPSET(0x679E, TAHITI)
CHIPSET(0x679F, TAHITI)

CHIPSET(0x6800, PITCAIRN)
CHIPSET(0x6801, PITCAIRN)
CHIPSET(0x6802, PITCAIRN)
CHIPSET(0x6806, PITCAIRN)
CHIPSET(0x6808, PITCAIRN)
CHIPSET(0x6809, PITCAIRN)
CHIPSET(0x6810, PITCAIRN)
CHIPSET(0x6811, PITCAIRN)
CHIPSET(0x6816, PITCAIRN)
CHIPSET(0x6817, PITCAIRN)
CHIPSET(0x6818, PITCAIRN)
CHIPSET(0x6819, PITCAIRN)
CHIPSET(0x684C, PITCAIRN)

CHIPSET(0x6820, VERDE)
CHIPSET(0x6821, VERDE)
CHIPSET(0x6822, VERDE)
CHIPSET(0x6823, VERDE)
CHIPSET(0x6824, VERDE)
CHIPSET(0x6825, VERDE)
CHIPSET(0x6826, VERDE)
CHIPSET(0x6827, VERDE)
CHIPSET(0x6828, VERDE)
CHIPSET(0x6829, VERDE)
CHIPSET(0x682A, VERDE)
CHIPSET(0x682B, VERDE)
CHIPSET(0x682C, VERDE)
CHIPSET(0x682D, VERDE)
CHIPSET(0x682F, VERDE)
CHIPSET(0x6830, VERDE)
CHIPSET(0x6831, VERDE)
CHIPSET(0x6835, VERDE)
CHIPSET(0x6837, VERDE)
CHIPSET(0x6838, VERDE)
CHIPSET(0x6839, VERDE)
CHIPSET(0x683B, VERDE)
CHIPSET(0x683D, VERDE)
CHIPSET(0x683F, VERDE)

CHIPSET(0x6600, OLAND)
CHIPSET(0x6601, OLAND)
CHIPSET(0x6602, OLAND)
CHIPSET(0x6603, OLAND)
CHIPSET(0x6604, OLAND)
CHIPSET(0x6605, OLAND)
CHIPSET(0x6606, OLAND)
CHIPSET(0x6607, OLAND)
CHIPSET(0x6608, OLAND)
CHIPSET(0x6610, OLAND)
CHIPSET(0x6611, OLAND)
CHIPSET(0x6613, OLAND)
CHIPSET(0x6617, OLAND)
CHIPSET(0x6620, OLAND)
CHIPSET(0x6621, OLAND)
CHIPSET(0x6623, OLAND)
CHIPSET(0x6631, OLAND)

CHIPSET(0x6660, HAINAN)
CHIPSET(0x6663, HAINAN)
CHIPSET(0x6664, HAINAN)
CHIPSET(0x6665, HAINAN)
CHIPSET(0x6667, HAINAN)
CHIPSET(0x666F, HAINAN)

CHIPSET(0x6640, BONAIRE)
CHIPSET(0x6641, BONAIRE)
CHIPSET(0x6646, BONAIRE)
CHIPSET(0x6647, BONAIRE)
CHIPSET(0x6649, BONAIRE)
CHIPSET(0x6650, BONAIRE)
CHIPSET(0x6651, BONAIRE)
CHIPSET(0x6658, BONAIRE)
CHIPSET(0x665C, BONAIRE)
CHIPSET(0x665D, BONAIRE)
CHIPSET(0x665F, BONAIRE)

CHIPSET(0x9830, KABINI)
CHIPSET(0x9831, KABINI)
CHIPSET(0x9832, KABINI)
CHIPSET(0x9833, KABINI)
CHIPSET(0x9834, KABINI)
CHIPSET(0x9835, KABINI)
CHIPSET(0x9836, KABINI)
CHIPSET(0x9837, KABINI)
CHIPSET(0x9838, KABINI)
CHIPSET(0x9839, KABINI)
CHIPSET(0x983A, KABINI)
CHIPSET(0x983B, KABINI)
CHIPSET(0x983C, KABINI)
CHIPSET(0x983D, KABINI)
CHIPSET(0x983E, KABINI)
CHIPSET(0x983F, KABINI)

CHIPSET(0x9850, KABINI)
CHIPSET(0x9851, KABINI)
CHIPSET(0x9852, KABINI)
CHIPSET(0x9853, KABINI)
CHIPSET(0x9854, KABINI)
CHIPSET(0x9855, KABINI)
CHIPSET(0x9856, KABINI)
CHIPSET(0x9857, KABINI)
CHIPSET(0x9858, KABINI)
CHIPSET(0x9859, KABINI)
CHIPSET(0x985A, KABINI)
CHIPSET(0x985B, KABINI)
CHIPSET(0x985C, KABINI)
CHIPSET(0x985D, KABINI)
CHIPSET(0x985E, KABINI)
CHIPSET(0x985F, KABINI)

CHIPSET(0x1304, KAVERI)
CHIPSET(0x1305, KAVERI)
CHIPSET(0x1306, KAVERI)
CHIPSET(0x1307, KAVERI)
CHIPSET(0x1309, KAVERI)
CHIPSET(0x130A, KAVERI)
CHIPSET(0x130B, KAVERI)
CHIPSET(0x130C, KAVERI)
CHIPSET(0x130D, KAVERI)
CHIPSET(0x130E, KAVERI)
CHIPSET(0x130F, KAVERI)
CHIPSET(0x1310, KAVERI)
CHIPSET(0x1311, KAVERI)
CHIPSET(0x1312, KAVERI)
CHIPSET(0x1313, KAVERI)
CHIPSET(0x1315, KAVERI)
CHIPSET(0x1316, KAVERI)
CHIPSET(0x1317, KAVERI)
CHIPSET(0x1318, KAVERI)
CHIPSET(0x131B, KAVERI)
CHIPSET(0x131C, KAVERI)
CHIPSET(0x131D, KAVERI)

CHIPSET(0x67A0, HAWAII)
CHIPSET(0x67A1, HAWAII)
CHIPSET(0x67A2, HAWAII)
CHIPSET(0x67A8, HAWAII)
CHIPSET(0x67A9, HAWAII)
CHIPSET(0x67AA, HAWAII)
CHIPSET(0x67B0, HAWAII)
CHIPSET(0x67B1, HAWAII)
CHIPSET(0x67B8, HAWAII)
CHIPSET(0x67B9, HAWAII)
CHIPSET(0x67BA, HAWAII)
CHIPSET(0x67BE, HAWAII)
