/*
 * Copyright © 2022 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/* Auto-generated file - don't edit */

#ifndef PVR_SIMPLE_COMPUTE_H
#define PVR_SIMPLE_COMPUTE_H

#include <stdint.h>

/* clang-format off */
static const uint8_t pvr_simple_compute_shader[] = {
   0x44, 0x12, 0xd3, 0x3f,
   0x00, 0x00, 0x00, 0x24,
   0x46, 0x40, 0xf9, 0xb0,
   0x87, 0x80, 0x40, 0xa0,
   0x2a, 0x30, 0x00, 0x02,
   0x04, 0x81, 0x61, 0x00,
   0x54, 0x00, 0x00, 0x00,
   0x02, 0x80, 0x6c, 0xc4,
   0x46, 0x40, 0xf9, 0xb0,
   0x87, 0x81, 0x40, 0xa0,
   0x2d, 0x10, 0x00, 0x02,
   0x46, 0x12, 0xd3, 0x3f,
   0x80, 0xca, 0x83, 0x10,
   0x00, 0x00, 0x25, 0xff,
   0x46, 0x13, 0xd3, 0x3f,
   0x80, 0xcb, 0x83, 0x10,
   0x00, 0x00, 0x25, 0xff,
   0x46, 0x40, 0xf9, 0xb0,
   0x87, 0x80, 0x40, 0xa0,
   0x25, 0x10, 0x00, 0x02,
   0x04, 0x81, 0x61, 0x00,
   0x14, 0x00, 0x00, 0x00,
   0x02, 0x80, 0x6c, 0x44,
   0x04, 0x80, 0x60, 0x00,
   0xc0, 0xff, 0xff, 0xff,
   0x02, 0x80, 0x6c, 0x04,
   0x89, 0x52, 0xdf, 0x3c,
   0xfc, 0xa0, 0x9c, 0x1e,
   0x87, 0x87, 0x80, 0xcf,
   0x90, 0x11, 0x01, 0xa0,
   0x25, 0xff, 0x46, 0x40,
   0xff, 0xd0, 0x87, 0xa5,
   0x40, 0xa0, 0x00, 0x10,
   0x00, 0x02, 0x44, 0x82,
   0x67, 0x38, 0x24, 0x00,
   0x24, 0xff, 0x04, 0x80,
   0x60, 0x04, 0x68, 0x00,
   0x00, 0x00, 0x45, 0x12,
   0xd3, 0x3f, 0xc0, 0x04,
   0x00, 0x00, 0x00, 0x25,
   0x27, 0x02, 0xeb, 0xa5,
   0x44, 0xa0, 0x00, 0x80,
   0x81, 0x08, 0x00, 0xc0,
   0x80, 0x04, 0x27, 0x04,
   0xeb, 0xa5, 0x44, 0xa0,
   0x00, 0x80, 0x81, 0x08,
   0x00, 0xc0, 0x81, 0x04,
   0x55, 0x20, 0xf1, 0x84,
   0x02, 0x00, 0x82, 0xc0,
   0x18, 0x00, 0x02, 0x80,
   0x6a, 0xff, 0x27, 0x02,
   0xeb, 0xa5, 0x44, 0xa0,
   0x02, 0x80, 0x83, 0x08,
   0x00, 0xc0, 0x80, 0x04,
   0x27, 0x04, 0xeb, 0xa5,
   0x44, 0xa0, 0x02, 0x80,
   0x83, 0x08, 0x00, 0xc0,
   0x81, 0x04, 0x66, 0x20,
   0xf2, 0x86, 0xb8, 0x28,
   0x00, 0x82, 0xc0, 0x18,
   0x00, 0xff, 0x02, 0x80,
   0x6a, 0xff, 0x45, 0x82,
   0x67, 0x24, 0x24, 0x00,
   0x24, 0xff, 0xf1, 0xff,
   0x04, 0x80, 0xee, 0x00,
   0xf2, 0xff, 0xff, 0xff
};
/* clang-format on */

#endif /* PVR_SIMPLE_COMPUTE_H */
