/*
 * Copyright © 2022 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef PVR_CDM_LOAD_SR_H
#define PVR_CDM_LOAD_SR_H

#include <stdint.h>

/* clang-format off */
static const uint8_t pvr_cdm_load_sr_code[] = {
	0x25, 0x02, 0x87, 0x81,
	0x04, 0x00, 0x00, 0x00,
	0x84, 0x04, 0x25, 0x02,
	0x87, 0x80, 0x04, 0x00,
	0x00, 0x00, 0x85, 0x04,
	0x25, 0x02, 0x87, 0x83,
	0x04, 0x00, 0x00, 0x00,
	0x86, 0x04, 0x25, 0x02,
	0x87, 0x82, 0x04, 0x00,
	0x00, 0x00, 0x87, 0x04,
	0x56, 0x20, 0xF1, 0x85,
	0x02, 0x80, 0x81, 0xD0,
	0xC4, 0x08, 0x00, 0xFF,
	0x02, 0x80, 0x6A, 0xFF,
	0x67, 0xF0, 0x40, 0x20,
	0x41, 0x8C, 0x80, 0x40,
	0x00, 0x50, 0x8F, 0xC0,
	0x80, 0x02, 0x04, 0x81,
	0x60, 0x00, 0x0A, 0x01,
	0x00, 0x00, 0x25, 0x36,
	0x87, 0x87, 0x00, 0x00,
	0x40, 0x05, 0xD1, 0x06,
	0x55, 0x20, 0xF1, 0x81,
	0x02, 0x00, 0xC0, 0xC6,
	0x08, 0x00, 0x02, 0x80,
	0x6A, 0xFF, 0x46, 0x42,
	0xD0, 0x03, 0xEA, 0xD1,
	0x41, 0x00, 0x01, 0x00,
	0x00, 0x51, 0x27, 0x06,
	0xEB, 0x84, 0x50, 0x20,
	0x86, 0x87, 0x04, 0x00,
	0xC0, 0x06, 0x87, 0x22,
	0x25, 0x32, 0x87, 0x87,
	0x00, 0x1F, 0x40, 0xC5,
	0x0C, 0xFF, 0x25, 0x02,
	0x87, 0xC0, 0x0C, 0x00,
	0x00, 0x00, 0x83, 0x0C,
	0x47, 0x42, 0xD0, 0x03,
	0xEA, 0x85, 0x41, 0x90,
	0x01, 0x08, 0x00, 0x00,
	0x85, 0x0C, 0x47, 0x42,
	0xD0, 0x03, 0xEA, 0xC5,
	0x41, 0x90, 0x01, 0x08,
	0x00, 0x00, 0xC5, 0x0C,
	0x67, 0xF0, 0x40, 0x28,
	0x42, 0x8C, 0x80, 0x40,
	0x80, 0xC5, 0x80, 0x90,
	0x80, 0xFF, 0x04, 0x81,
	0x60, 0x00, 0xCC, 0xFF,
	0xFF, 0xFF, 0x66, 0xF0,
	0x40, 0x28, 0x42, 0x8C,
	0x80, 0x40, 0x00, 0x51,
	0xD0, 0x80, 0x04, 0x81,
	0x60, 0x00, 0x86, 0xFF,
	0xFF, 0xFF, 0x46, 0x40,
	0xF1, 0xB0, 0xE2, 0x81,
	0x4D, 0x01, 0x00, 0x00,
	0x00, 0xFF, 0x04, 0x81,
	0x60, 0x00, 0x32, 0x00,
	0x00, 0x00, 0x02, 0x80,
	0x6C, 0xC4, 0x45, 0x12,
	0xD3, 0x3F, 0x01, 0x00,
	0x00, 0xAB, 0x01, 0xFF,
	0x44, 0x12, 0xD3, 0x3F,
	0x00, 0x00, 0x00, 0x40,
	0x44, 0x10, 0xD3, 0x3F,
	0x40, 0x00, 0x00, 0xFF,
	0x02, 0x80, 0x6C, 0x84,
	0x04, 0x80, 0x60, 0x00,
	0x40, 0x00, 0x00, 0x00,
	0x46, 0x50, 0xFB, 0xB0,
	0x87, 0xE2, 0x81, 0x4B,
	0x03, 0x00, 0x00, 0x00,
	0x04, 0x81, 0x61, 0x00,
	0xF4, 0xFF, 0xFF, 0xFF,
	0x02, 0x80, 0x6C, 0xC4,
	0x45, 0x12, 0xD3, 0x3F,
	0x01, 0x00, 0x00, 0xAA,
	0x01, 0xFF, 0x44, 0x12,
	0xD3, 0x3F, 0x00, 0x00,
	0x00, 0x40, 0x44, 0x10,
	0xD3, 0x3F, 0x40, 0x00,
	0x00, 0xFF, 0x03, 0x80,
	0x6C, 0x84, 0xF1, 0xFF,
	0x04, 0x80, 0xEE, 0x00,
	0xF2, 0xFF, 0xFF, 0xFF
};
/* clang-format on */

#endif /* PVR_CDM_LOAD_SR_H */
