#ifndef _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NV01_2D_XML
#define _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NV01_2D_XML

/* Autogenerated file, DO NOT EDIT manually!

This file was generated by the rules-ng-ng headergen tool in this git repository:
http://0x04.net/cgit/index.cgi/rules-ng-ng
git clone git://0x04.net/rules-ng-ng

The rules-ng-ng source files this header was generated from are:
- /home/skeggsb/git/envytools/rnndb/nv_objects.xml      (    794 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/copyright.xml       (   6452 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_m2mf.xml         (   2696 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_object.xml       (  12672 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvchipsets.xml      (   3617 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_defs.xml         (   4437 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_defs.xml       (   5468 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvc0_m2mf.xml       (   2687 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv01_2d.xml         (  32584 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv04_dvd.xml        (   3000 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv03_3d.xml         (   5209 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv04_3d.xml         (  17759 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_3ddefs.xml       (  16394 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv10_3d.xml         (  18437 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv20_3d.xml         (  21107 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv30-40_3d.xml      (  31987 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_2d.xml         (  11113 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_3d.xml         (  65233 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nv50_compute.xml    (  14012 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv84_crypt.xml      (   2071 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nv31_mpeg.xml       (   2269 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvc0_3d.xml         (  52547 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nvc0_compute.xml    (  10865 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/blob_nvc0_pcopy.xml (   4516 bytes, from 2011-10-22 08:01:09)

Copyright (C) 2006-2011 by the following authors:
- Artur Huillet <arthur.huillet@free.fr> (ahuillet)
- Ben Skeggs (darktama, darktama_)
- B. R. <koala_br@users.sourceforge.net> (koala_br)
- Carlos Martin <carlosmn@users.sf.net> (carlosmn)
- Christoph Bumiller <e0425955@student.tuwien.ac.at> (calim, chrisbmr)
- Dawid Gajownik <gajownik@users.sf.net> (gajownik)
- Dmitry Baryshkov
- Dmitry Eremin-Solenikov <lumag@users.sf.net> (lumag)
- EdB <edb_@users.sf.net> (edb_)
- Erik Waling <erikwailing@users.sf.net> (erikwaling)
- Francisco Jerez <currojerez@riseup.net> (curro)
- imirkin <imirkin@users.sf.net> (imirkin)
- jb17bsome <jb17bsome@bellsouth.net> (jb17bsome)
- Jeremy Kolb <kjeremy@users.sf.net> (kjeremy)
- Laurent Carlier <lordheavym@gmail.com> (lordheavy)
- Luca Barbieri <luca@luca-barbieri.com> (lb, lb1)
- Maarten Maathuis <madman2003@gmail.com> (stillunknown)
- Marcin Kościelnicki <koriakin@0x04.net> (mwk, koriakin)
- Mark Carey <mark.carey@gmail.com> (careym)
- Matthieu Castet <matthieu.castet@parrot.com> (mat-c)
- nvidiaman <nvidiaman@users.sf.net> (nvidiaman)
- Patrice Mandin <patmandin@gmail.com> (pmandin, pmdata)
- Pekka Paalanen <pq@iki.fi> (pq, ppaalanen)
- Peter Popov <ironpeter@users.sf.net> (ironpeter)
- Richard Hughes <hughsient@users.sf.net> (hughsient)
- Rudi Cilibrasi <cilibrar@users.sf.net> (cilibrar)
- Serge Martin
- Simon Raffeiner
- Stephane Loeuillet <leroutier@users.sf.net> (leroutier)
- Stephane Marchesin <stephane.marchesin@gmail.com> (marcheu)
- sturmflut <sturmflut@users.sf.net> (sturmflut)
- Sylvain Munaut <tnt@246tNt.com>
- Victor Stinner <victor.stinner@haypocalc.com> (haypo)
- Wladmir van der Laan <laanwj@gmail.com> (miathan6)
- Younes Manton <younes.m@gmail.com> (ymanton)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/



#define NV01_BETA_DMA_NOTIFY					0x00000180

#define NV01_BETA_BETA_1D31					0x00000300


#define NV04_BETA4_DMA_NOTIFY					0x00000180

#define NV04_BETA4_BETA_FACTOR					0x00000300


#define NV01_CHROMA_DMA_NOTIFY					0x00000180

#define NV01_CHROMA_COLOR_FORMAT				0x00000300
#define NV01_CHROMA_COLOR_FORMAT_A16R5G6B5			0x00000001
#define NV01_CHROMA_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV01_CHROMA_COLOR_FORMAT_A8R8G8B8			0x00000003

#define NV01_CHROMA_COLOR					0x00000304


#define NV01_PATTERN_DMA_NOTIFY					0x00000180

#define NV01_PATTERN_COLOR_FORMAT				0x00000300
#define NV01_PATTERN_COLOR_FORMAT_A16R5G6B5			0x00000001
#define NV01_PATTERN_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV01_PATTERN_COLOR_FORMAT_A8R8G8B8			0x00000003

#define NV01_PATTERN_MONOCHROME_FORMAT				0x00000304
#define NV01_PATTERN_MONOCHROME_FORMAT_CGA6			0x00000001
#define NV01_PATTERN_MONOCHROME_FORMAT_LE			0x00000002

#define NV01_PATTERN_MONOCHROME_SHAPE				0x00000308
#define NV01_PATTERN_MONOCHROME_SHAPE_8X8			0x00000000
#define NV01_PATTERN_MONOCHROME_SHAPE_64X1			0x00000001
#define NV01_PATTERN_MONOCHROME_SHAPE_1X64			0x00000002

#define NV04_PATTERN_PATTERN_SELECT				0x0000030c
#define NV04_PATTERN_PATTERN_SELECT_MONO			0x00000001
#define NV04_PATTERN_PATTERN_SELECT_COLOR			0x00000002

#define NV01_PATTERN_MONOCHROME_COLOR(i0)		       (0x00000310 + 0x4*(i0))
#define NV01_PATTERN_MONOCHROME_COLOR__ESIZE			0x00000004
#define NV01_PATTERN_MONOCHROME_COLOR__LEN			0x00000002

#define NV01_PATTERN_MONOCHROME_PATTERN(i0)		       (0x00000318 + 0x4*(i0))
#define NV01_PATTERN_MONOCHROME_PATTERN__ESIZE			0x00000004
#define NV01_PATTERN_MONOCHROME_PATTERN__LEN			0x00000002

#define NV04_PATTERN_PATTERN_Y8(i0)			       (0x00000400 + 0x4*(i0))
#define NV04_PATTERN_PATTERN_Y8__ESIZE				0x00000004
#define NV04_PATTERN_PATTERN_Y8__LEN				0x00000010
#define NV04_PATTERN_PATTERN_Y8_Y0__MASK			0x000000ff
#define NV04_PATTERN_PATTERN_Y8_Y0__SHIFT			0
#define NV04_PATTERN_PATTERN_Y8_Y1__MASK			0x0000ff00
#define NV04_PATTERN_PATTERN_Y8_Y1__SHIFT			8
#define NV04_PATTERN_PATTERN_Y8_Y2__MASK			0x00ff0000
#define NV04_PATTERN_PATTERN_Y8_Y2__SHIFT			16
#define NV04_PATTERN_PATTERN_Y8_Y3__MASK			0xff000000
#define NV04_PATTERN_PATTERN_Y8_Y3__SHIFT			24

#define NV04_PATTERN_PATTERN_R5G6B5(i0)			       (0x00000500 + 0x4*(i0))
#define NV04_PATTERN_PATTERN_R5G6B5__ESIZE			0x00000004
#define NV04_PATTERN_PATTERN_R5G6B5__LEN			0x00000020
#define NV04_PATTERN_PATTERN_R5G6B5_B0__MASK			0x0000001f
#define NV04_PATTERN_PATTERN_R5G6B5_B0__SHIFT			0
#define NV04_PATTERN_PATTERN_R5G6B5_G0__MASK			0x000007e0
#define NV04_PATTERN_PATTERN_R5G6B5_G0__SHIFT			5
#define NV04_PATTERN_PATTERN_R5G6B5_R0__MASK			0x0000f800
#define NV04_PATTERN_PATTERN_R5G6B5_R0__SHIFT			11
#define NV04_PATTERN_PATTERN_R5G6B5_B1__MASK			0x001f0000
#define NV04_PATTERN_PATTERN_R5G6B5_B1__SHIFT			16
#define NV04_PATTERN_PATTERN_R5G6B5_G1__MASK			0x07e00000
#define NV04_PATTERN_PATTERN_R5G6B5_G1__SHIFT			21
#define NV04_PATTERN_PATTERN_R5G6B5_R1__MASK			0xf8000000
#define NV04_PATTERN_PATTERN_R5G6B5_R1__SHIFT			27

#define NV04_PATTERN_PATTERN_X1R5G5B5(i0)		       (0x00000600 + 0x4*(i0))
#define NV04_PATTERN_PATTERN_X1R5G5B5__ESIZE			0x00000004
#define NV04_PATTERN_PATTERN_X1R5G5B5__LEN			0x00000020
#define NV04_PATTERN_PATTERN_X1R5G5B5_B0__MASK			0x0000001f
#define NV04_PATTERN_PATTERN_X1R5G5B5_B0__SHIFT			0
#define NV04_PATTERN_PATTERN_X1R5G5B5_G0__MASK			0x000003e0
#define NV04_PATTERN_PATTERN_X1R5G5B5_G0__SHIFT			5
#define NV04_PATTERN_PATTERN_X1R5G5B5_R0__MASK			0x00007c00
#define NV04_PATTERN_PATTERN_X1R5G5B5_R0__SHIFT			10
#define NV04_PATTERN_PATTERN_X1R5G5B5_B1__MASK			0x001f0000
#define NV04_PATTERN_PATTERN_X1R5G5B5_B1__SHIFT			16
#define NV04_PATTERN_PATTERN_X1R5G5B5_G1__MASK			0x03e00000
#define NV04_PATTERN_PATTERN_X1R5G5B5_G1__SHIFT			21
#define NV04_PATTERN_PATTERN_X1R5G5B5_R1__MASK			0x7c000000
#define NV04_PATTERN_PATTERN_X1R5G5B5_R1__SHIFT			26

#define NV04_PATTERN_PATTERN_X8R8G8B8(i0)		       (0x00000700 + 0x4*(i0))
#define NV04_PATTERN_PATTERN_X8R8G8B8__ESIZE			0x00000004
#define NV04_PATTERN_PATTERN_X8R8G8B8__LEN			0x00000040
#define NV04_PATTERN_PATTERN_X8R8G8B8_B__MASK			0x000000ff
#define NV04_PATTERN_PATTERN_X8R8G8B8_B__SHIFT			0
#define NV04_PATTERN_PATTERN_X8R8G8B8_G__MASK			0x0000ff00
#define NV04_PATTERN_PATTERN_X8R8G8B8_G__SHIFT			8
#define NV04_PATTERN_PATTERN_X8R8G8B8_R__MASK			0x00ff0000
#define NV04_PATTERN_PATTERN_X8R8G8B8_R__SHIFT			16


#define NV01_CLIP_DMA_NOTIFY					0x00000180

#define NV01_CLIP_POINT						0x00000300
#define NV01_CLIP_POINT_X__MASK					0x0000ffff
#define NV01_CLIP_POINT_X__SHIFT				0
#define NV01_CLIP_POINT_Y__MASK					0xffff0000
#define NV01_CLIP_POINT_Y__SHIFT				16

#define NV01_CLIP_SIZE						0x00000304
#define NV01_CLIP_SIZE_W__MASK					0x0000ffff
#define NV01_CLIP_SIZE_W__SHIFT					0
#define NV01_CLIP_SIZE_H__MASK					0xffff0000
#define NV01_CLIP_SIZE_H__SHIFT					16


#define NV01_ROP_DMA_NOTIFY					0x00000180

#define NV01_ROP_ROP						0x00000300


#define NV04_SURFACE_2D_DMA_NOTIFY				0x00000180

#define NV04_SURFACE_2D_DMA_IMAGE_SOURCE			0x00000184

#define NV04_SURFACE_2D_DMA_IMAGE_DESTIN			0x00000188


#define NV50_SURFACE_2D_SRC_LINEAR				0x00000200

#define NV50_SURFACE_2D_SRC_TILE_MODE				0x00000204

#define NV50_SURFACE_2D_SRC_WIDTH				0x00000208

#define NV50_SURFACE_2D_SRC_HEIGHT				0x0000020c

#define NV50_SURFACE_2D_UNK0210					0x00000210

#define NV50_SURFACE_2D_UNK0214					0x00000214

#define NV50_SURFACE_2D_DST_LINEAR				0x00000218

#define NV50_SURFACE_2D_DST_TILE_MODE				0x0000021c

#define NV50_SURFACE_2D_DST_WIDTH				0x00000220

#define NV50_SURFACE_2D_DST_HEIGHT				0x00000224

#define NV50_SURFACE_2D_UNK0228					0x00000228

#define NV50_SURFACE_2D_UNK022C					0x0000022c

#define NV50_SURFACE_2D_OFFSET_SOURCE_HIGH			0x00000230

#define NV50_SURFACE_2D_OFFSET_DESTIN_HIGH			0x00000234

#define NV04_SURFACE_2D_FORMAT					0x00000300
#define NV04_SURFACE_2D_FORMAT_Y8				0x00000001
#define NV04_SURFACE_2D_FORMAT_X1R5G5B5_Z1R5G5B5		0x00000002
#define NV04_SURFACE_2D_FORMAT_X1R5G5B5_X1R5G5B5		0x00000003
#define NV04_SURFACE_2D_FORMAT_R5G6B5				0x00000004
#define NV04_SURFACE_2D_FORMAT_Y16				0x00000005
#define NV04_SURFACE_2D_FORMAT_X8R8G8B8_Z8R8G8B8		0x00000006
#define NV04_SURFACE_2D_FORMAT_X8R8G8B8_X8R8G8B8		0x00000007
#define NV04_SURFACE_2D_FORMAT_X1A7R8G8B8_Z1A7R8G8B8		0x00000008
#define NV04_SURFACE_2D_FORMAT_X1A7R8G8B8_X1A7R8G8B8		0x00000009
#define NV04_SURFACE_2D_FORMAT_A8R8G8B8				0x0000000a
#define NV04_SURFACE_2D_FORMAT_Y32				0x0000000b

#define NV04_SURFACE_2D_PITCH					0x00000304
#define NV04_SURFACE_2D_PITCH_SOURCE__MASK			0x0000ffff
#define NV04_SURFACE_2D_PITCH_SOURCE__SHIFT			0
#define NV04_SURFACE_2D_PITCH_DESTIN__MASK			0xffff0000
#define NV04_SURFACE_2D_PITCH_DESTIN__SHIFT			16

#define NV04_SURFACE_2D_OFFSET_SOURCE				0x00000308

#define NV04_SURFACE_2D_OFFSET_DESTIN				0x0000030c


#define NV04_SURFACE_SWZ_DMA_NOTIFY				0x00000180

#define NV04_SURFACE_SWZ_DMA_IMAGE				0x00000184

#define NV04_SURFACE_SWZ_FORMAT					0x00000300
#define NV04_SURFACE_SWZ_FORMAT_COLOR__MASK			0x000000ff
#define NV04_SURFACE_SWZ_FORMAT_COLOR__SHIFT			0
#define NV04_SURFACE_SWZ_FORMAT_COLOR_Y8			0x00000001
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X1R5G5B5_Z1R5G5B5		0x00000002
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X1R5G5B5_X1R5G5B5		0x00000003
#define NV04_SURFACE_SWZ_FORMAT_COLOR_R5G6B5			0x00000004
#define NV04_SURFACE_SWZ_FORMAT_COLOR_Y16			0x00000005
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X8R8G8B8_Z8R8G8B8		0x00000006
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X8R8G8B8_X8R8G8B8		0x00000007
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X1A7R8G8B8_Z1A7R8G8B8	0x00000008
#define NV04_SURFACE_SWZ_FORMAT_COLOR_X1A7R8G8B8_X1A7R8G8B8	0x00000009
#define NV04_SURFACE_SWZ_FORMAT_COLOR_A8R8G8B8			0x0000000a
#define NV04_SURFACE_SWZ_FORMAT_COLOR_Y32			0x0000000b
#define NV04_SURFACE_SWZ_FORMAT_BASE_SIZE_U__MASK		0x00ff0000
#define NV04_SURFACE_SWZ_FORMAT_BASE_SIZE_U__SHIFT		16
#define NV04_SURFACE_SWZ_FORMAT_BASE_SIZE_V__MASK		0xff000000
#define NV04_SURFACE_SWZ_FORMAT_BASE_SIZE_V__SHIFT		24

#define NV04_SURFACE_SWZ_OFFSET					0x00000304


#define NV01_POINT_PATCH					0x0000010c

#define NV01_POINT_DMA_NOTIFY					0x00000180

#define NV01_POINT_CLIP						0x00000184

#define NV01_POINT_PATTERN					0x00000188

#define NV01_POINT_ROP						0x0000018c

#define NV01_POINT_BETA						0x00000190

#define NV01_POINT_SURFACE_DST					0x00000194

#define NV01_POINT_OPERATION					0x000002fc
#define NV01_POINT_OPERATION_SRCCOPY_AND			0x00000000
#define NV01_POINT_OPERATION_ROP_AND				0x00000001
#define NV01_POINT_OPERATION_BLEND_AND				0x00000002
#define NV01_POINT_OPERATION_SRCCOPY				0x00000003
#define NV01_POINT_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_POINT_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_POINT_COLOR_FORMAT					0x00000300
#define NV01_POINT_COLOR_FORMAT_X16A8Y8				0x00000001
#define NV01_POINT_COLOR_FORMAT_X24Y8				0x00000002
#define NV01_POINT_COLOR_FORMAT_X16A1R5G5B5			0x00000003
#define NV01_POINT_COLOR_FORMAT_X17R5G5B5			0x00000004
#define NV01_POINT_COLOR_FORMAT_A8R8G8B8			0x00000005
#define NV01_POINT_COLOR_FORMAT_X8R8G8B8			0x00000006
#define NV01_POINT_COLOR_FORMAT_A16Y16				0x00000007
#define NV01_POINT_COLOR_FORMAT_X16Y16				0x00000008

#define NV01_POINT_COLOR					0x00000304

#define NV01_POINT_POINT(i0)				       (0x00000400 + 0x4*(i0))
#define NV01_POINT_POINT__ESIZE					0x00000004
#define NV01_POINT_POINT__LEN					0x00000020
#define NV01_POINT_POINT_X__MASK				0x0000ffff
#define NV01_POINT_POINT_X__SHIFT				0
#define NV01_POINT_POINT_Y__MASK				0xffff0000
#define NV01_POINT_POINT_Y__SHIFT				16

#define NV01_POINT_POINT32_X(i0)			       (0x00000480 + 0x8*(i0))
#define NV01_POINT_POINT32_X__ESIZE				0x00000008
#define NV01_POINT_POINT32_X__LEN				0x00000010

#define NV01_POINT_POINT32_Y(i0)			       (0x00000484 + 0x8*(i0))
#define NV01_POINT_POINT32_Y__ESIZE				0x00000008
#define NV01_POINT_POINT32_Y__LEN				0x00000010

#define NV01_POINT_CPOINT_COLOR(i0)			       (0x00000500 + 0x8*(i0))
#define NV01_POINT_CPOINT_COLOR__ESIZE				0x00000008
#define NV01_POINT_CPOINT_COLOR__LEN				0x00000010

#define NV01_POINT_CPOINT_POINT(i0)			       (0x00000504 + 0x8*(i0))
#define NV01_POINT_CPOINT_POINT__ESIZE				0x00000008
#define NV01_POINT_CPOINT_POINT__LEN				0x00000010
#define NV01_POINT_CPOINT_POINT_X__MASK				0x0000ffff
#define NV01_POINT_CPOINT_POINT_X__SHIFT			0
#define NV01_POINT_CPOINT_POINT_Y__MASK				0xffff0000
#define NV01_POINT_CPOINT_POINT_Y__SHIFT			16


#define NV01_LINE_PATCH						0x0000010c

#define NV01_LINE_DMA_NOTIFY					0x00000180

#define NV01_LINE_CLIP						0x00000184

#define NV01_LINE_PATTERN					0x00000188

#define NV04_LIN_PATTERN					0x00000188

#define NV01_LINE_ROP						0x0000018c

#define NV01_LINE_BETA						0x00000190

#define NV01_LINE_SURFACE_DST					0x00000194


#define NV04_LIN_BETA4						0x00000194

#define NV04_LIN_SURFACE					0x00000198

#define NV01_LINE_OPERATION					0x000002fc
#define NV01_LINE_OPERATION_SRCCOPY_AND				0x00000000
#define NV01_LINE_OPERATION_ROP_AND				0x00000001
#define NV01_LINE_OPERATION_BLEND_AND				0x00000002
#define NV01_LINE_OPERATION_SRCCOPY				0x00000003
#define NV01_LINE_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_LINE_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_LINE_COLOR_FORMAT					0x00000300
#define NV01_LINE_COLOR_FORMAT_A16R5G6B5			0x00000001
#define NV01_LINE_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV01_LINE_COLOR_FORMAT_A8R8G8B8				0x00000003

#define NV01_LINE_COLOR						0x00000304

#define NV01_LINE_LINE_POINT0(i0)			       (0x00000400 + 0x8*(i0))
#define NV01_LINE_LINE_POINT0__ESIZE				0x00000008
#define NV01_LINE_LINE_POINT0__LEN				0x00000010
#define NV01_LINE_LINE_POINT0_X__MASK				0x0000ffff
#define NV01_LINE_LINE_POINT0_X__SHIFT				0
#define NV01_LINE_LINE_POINT0_Y__MASK				0xffff0000
#define NV01_LINE_LINE_POINT0_Y__SHIFT				16

#define NV01_LINE_LINE_POINT1(i0)			       (0x00000404 + 0x8*(i0))
#define NV01_LINE_LINE_POINT1__ESIZE				0x00000008
#define NV01_LINE_LINE_POINT1__LEN				0x00000010
#define NV01_LINE_LINE_POINT1_X__MASK				0x0000ffff
#define NV01_LINE_LINE_POINT1_X__SHIFT				0
#define NV01_LINE_LINE_POINT1_Y__MASK				0xffff0000
#define NV01_LINE_LINE_POINT1_Y__SHIFT				16

#define NV01_LINE_LINE32_POINT0_X(i0)			       (0x00000480 + 0x10*(i0))
#define NV01_LINE_LINE32_POINT0_X__ESIZE			0x00000010
#define NV01_LINE_LINE32_POINT0_X__LEN				0x00000008

#define NV01_LINE_LINE32_POINT0_Y(i0)			       (0x00000484 + 0x10*(i0))
#define NV01_LINE_LINE32_POINT0_Y__ESIZE			0x00000010
#define NV01_LINE_LINE32_POINT0_Y__LEN				0x00000008

#define NV01_LINE_LINE32_POINT1_X(i0)			       (0x00000488 + 0x10*(i0))
#define NV01_LINE_LINE32_POINT1_X__ESIZE			0x00000010
#define NV01_LINE_LINE32_POINT1_X__LEN				0x00000008

#define NV01_LINE_LINE32_POINT1_Y(i0)			       (0x0000048c + 0x10*(i0))
#define NV01_LINE_LINE32_POINT1_Y__ESIZE			0x00000010
#define NV01_LINE_LINE32_POINT1_Y__LEN				0x00000008

#define NV01_LINE_POLYLINE(i0)				       (0x00000500 + 0x4*(i0))
#define NV01_LINE_POLYLINE__ESIZE				0x00000004
#define NV01_LINE_POLYLINE__LEN					0x00000020
#define NV01_LINE_POLYLINE_X__MASK				0x0000ffff
#define NV01_LINE_POLYLINE_X__SHIFT				0
#define NV01_LINE_POLYLINE_Y__MASK				0xffff0000
#define NV01_LINE_POLYLINE_Y__SHIFT				16

#define NV01_LINE_POLYLINE32_POINT_X(i0)		       (0x00000580 + 0x8*(i0))
#define NV01_LINE_POLYLINE32_POINT_X__ESIZE			0x00000008
#define NV01_LINE_POLYLINE32_POINT_X__LEN			0x00000010

#define NV01_LINE_POLYLINE32_POINT_Y(i0)		       (0x00000584 + 0x8*(i0))
#define NV01_LINE_POLYLINE32_POINT_Y__ESIZE			0x00000008
#define NV01_LINE_POLYLINE32_POINT_Y__LEN			0x00000010

#define NV01_LINE_CPOLYLINE_COLOR(i0)			       (0x00000600 + 0x8*(i0))
#define NV01_LINE_CPOLYLINE_COLOR__ESIZE			0x00000008
#define NV01_LINE_CPOLYLINE_COLOR__LEN				0x00000010

#define NV01_LINE_CPOLYLINE_POINT(i0)			       (0x00000604 + 0x8*(i0))
#define NV01_LINE_CPOLYLINE_POINT__ESIZE			0x00000008
#define NV01_LINE_CPOLYLINE_POINT__LEN				0x00000010
#define NV01_LINE_CPOLYLINE_POINT_X__MASK			0x0000ffff
#define NV01_LINE_CPOLYLINE_POINT_X__SHIFT			0
#define NV01_LINE_CPOLYLINE_POINT_Y__MASK			0xffff0000
#define NV01_LINE_CPOLYLINE_POINT_Y__SHIFT			16


#define NV01_TRI_PATCH						0x0000010c

#define NV01_TRI_DMA_NOTIFY					0x00000180

#define NV01_TRI_CLIP						0x00000184

#define NV01_TRI_PATTERN					0x00000188

#define NV04_TRI_PATTERN					0x00000188

#define NV01_TRI_ROP						0x0000018c

#define NV01_TRI_BETA						0x00000190

#define NV01_TRI_SURFACE_DST					0x00000194


#define NV04_TRI_BETA4						0x00000194

#define NV04_TRI_SURFACE					0x00000198

#define NV01_TRI_OPERATION					0x000002fc
#define NV01_TRI_OPERATION_SRCCOPY_AND				0x00000000
#define NV01_TRI_OPERATION_ROP_AND				0x00000001
#define NV01_TRI_OPERATION_BLEND_AND				0x00000002
#define NV01_TRI_OPERATION_SRCCOPY				0x00000003
#define NV01_TRI_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_TRI_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_TRI_COLOR_FORMAT					0x00000300
#define NV01_TRI_COLOR_FORMAT_A16R5G6B5				0x00000001
#define NV01_TRI_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV01_TRI_COLOR_FORMAT_A8R8G8B8				0x00000003

#define NV01_TRI_COLOR						0x00000304

#define NV01_TRI_TRIANGLE_POINT0				0x00000310
#define NV01_TRI_TRIANGLE_POINT0_X__MASK			0x0000ffff
#define NV01_TRI_TRIANGLE_POINT0_X__SHIFT			0
#define NV01_TRI_TRIANGLE_POINT0_Y__MASK			0xffff0000
#define NV01_TRI_TRIANGLE_POINT0_Y__SHIFT			16

#define NV01_TRI_TRIANGLE_POINT1				0x00000314
#define NV01_TRI_TRIANGLE_POINT1_X__MASK			0x0000ffff
#define NV01_TRI_TRIANGLE_POINT1_X__SHIFT			0
#define NV01_TRI_TRIANGLE_POINT1_Y__MASK			0xffff0000
#define NV01_TRI_TRIANGLE_POINT1_Y__SHIFT			16

#define NV01_TRI_TRIANGLE_POINT2				0x00000318
#define NV01_TRI_TRIANGLE_POINT2_X__MASK			0x0000ffff
#define NV01_TRI_TRIANGLE_POINT2_X__SHIFT			0
#define NV01_TRI_TRIANGLE_POINT2_Y__MASK			0xffff0000
#define NV01_TRI_TRIANGLE_POINT2_Y__SHIFT			16

#define NV01_TRI_TRIANGLE32_POINT0_X				0x00000320

#define NV01_TRI_TRIANGLE32_POINT0_Y				0x00000324

#define NV01_TRI_TRIANGLE32_POINT1_X				0x00000328

#define NV01_TRI_TRIANGLE32_POINT1_Y				0x0000032c

#define NV01_TRI_TRIANGLE32_POINT2_X				0x00000330

#define NV01_TRI_TRIANGLE32_POINT2_Y				0x00000334

#define NV01_TRI_TRIMESH(i0)				       (0x00000400 + 0x4*(i0))
#define NV01_TRI_TRIMESH__ESIZE					0x00000004
#define NV01_TRI_TRIMESH__LEN					0x00000020
#define NV01_TRI_TRIMESH_X__MASK				0x0000ffff
#define NV01_TRI_TRIMESH_X__SHIFT				0
#define NV01_TRI_TRIMESH_Y__MASK				0xffff0000
#define NV01_TRI_TRIMESH_Y__SHIFT				16

#define NV01_TRI_TRIMESH32_POINT_X(i0)			       (0x00000480 + 0x8*(i0))
#define NV01_TRI_TRIMESH32_POINT_X__ESIZE			0x00000008
#define NV01_TRI_TRIMESH32_POINT_X__LEN				0x00000010

#define NV01_TRI_TRIMESH32_POINT_Y(i0)			       (0x00000484 + 0x8*(i0))
#define NV01_TRI_TRIMESH32_POINT_Y__ESIZE			0x00000008
#define NV01_TRI_TRIMESH32_POINT_Y__LEN				0x00000010

#define NV01_TRI_CTRIANGLE_COLOR(i0)			       (0x00000500 + 0x10*(i0))
#define NV01_TRI_CTRIANGLE_COLOR__ESIZE				0x00000010
#define NV01_TRI_CTRIANGLE_COLOR__LEN				0x00000008

#define NV01_TRI_CTRIANGLE_POINT0(i0)			       (0x00000504 + 0x10*(i0))
#define NV01_TRI_CTRIANGLE_POINT0__ESIZE			0x00000010
#define NV01_TRI_CTRIANGLE_POINT0__LEN				0x00000008
#define NV01_TRI_CTRIANGLE_POINT0_X__MASK			0x0000ffff
#define NV01_TRI_CTRIANGLE_POINT0_X__SHIFT			0
#define NV01_TRI_CTRIANGLE_POINT0_Y__MASK			0xffff0000
#define NV01_TRI_CTRIANGLE_POINT0_Y__SHIFT			16

#define NV01_TRI_CTRIANGLE_POINT1(i0)			       (0x00000508 + 0x10*(i0))
#define NV01_TRI_CTRIANGLE_POINT1__ESIZE			0x00000010
#define NV01_TRI_CTRIANGLE_POINT1__LEN				0x00000008
#define NV01_TRI_CTRIANGLE_POINT1_X__MASK			0x0000ffff
#define NV01_TRI_CTRIANGLE_POINT1_X__SHIFT			0
#define NV01_TRI_CTRIANGLE_POINT1_Y__MASK			0xffff0000
#define NV01_TRI_CTRIANGLE_POINT1_Y__SHIFT			16

#define NV01_TRI_CTRIANGLE_POINT2(i0)			       (0x0000050c + 0x10*(i0))
#define NV01_TRI_CTRIANGLE_POINT2__ESIZE			0x00000010
#define NV01_TRI_CTRIANGLE_POINT2__LEN				0x00000008
#define NV01_TRI_CTRIANGLE_POINT2_X__MASK			0x0000ffff
#define NV01_TRI_CTRIANGLE_POINT2_X__SHIFT			0
#define NV01_TRI_CTRIANGLE_POINT2_Y__MASK			0xffff0000
#define NV01_TRI_CTRIANGLE_POINT2_Y__SHIFT			16

#define NV01_TRI_CTRIMESH_COLOR(i0)			       (0x00000580 + 0x8*(i0))
#define NV01_TRI_CTRIMESH_COLOR__ESIZE				0x00000008
#define NV01_TRI_CTRIMESH_COLOR__LEN				0x00000010

#define NV01_TRI_CTRIMESH_POINT(i0)			       (0x00000584 + 0x8*(i0))
#define NV01_TRI_CTRIMESH_POINT__ESIZE				0x00000008
#define NV01_TRI_CTRIMESH_POINT__LEN				0x00000010
#define NV01_TRI_CTRIMESH_POINT_X__MASK				0x0000ffff
#define NV01_TRI_CTRIMESH_POINT_X__SHIFT			0
#define NV01_TRI_CTRIMESH_POINT_Y__MASK				0xffff0000
#define NV01_TRI_CTRIMESH_POINT_Y__SHIFT			16


#define NV01_RECT_PATCH						0x0000010c

#define NV01_RECT_DMA_NOTIFY					0x00000180

#define NV01_RECT_CLIP						0x00000184

#define NV01_RECT_PATTERN					0x00000188

#define NV04_RECT_PATTERN					0x00000188

#define NV01_RECT_ROP						0x0000018c

#define NV01_RECT_BETA						0x00000190

#define NV01_RECT_SURFACE_DST					0x00000194


#define NV04_RECT_BETA4						0x00000194

#define NV04_RECT_SURFACE					0x00000198

#define NV01_RECT_OPERATION					0x000002fc
#define NV01_RECT_OPERATION_SRCCOPY_AND				0x00000000
#define NV01_RECT_OPERATION_ROP_AND				0x00000001
#define NV01_RECT_OPERATION_BLEND_AND				0x00000002
#define NV01_RECT_OPERATION_SRCCOPY				0x00000003
#define NV01_RECT_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_RECT_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_RECT_COLOR_FORMAT					0x00000300
#define NV01_RECT_COLOR_FORMAT_A16R5G6B5			0x00000001
#define NV01_RECT_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV01_RECT_COLOR_FORMAT_A8R8G8B8				0x00000003

#define NV01_RECT_COLOR						0x00000304

#define NV01_RECT_RECTANGLE_POINT(i0)			       (0x00000400 + 0x8*(i0))
#define NV01_RECT_RECTANGLE_POINT__ESIZE			0x00000008
#define NV01_RECT_RECTANGLE_POINT__LEN				0x00000010
#define NV01_RECT_RECTANGLE_POINT_X__MASK			0x0000ffff
#define NV01_RECT_RECTANGLE_POINT_X__SHIFT			0
#define NV01_RECT_RECTANGLE_POINT_Y__MASK			0xffff0000
#define NV01_RECT_RECTANGLE_POINT_Y__SHIFT			16

#define NV01_RECT_RECTANGLE_SIZE(i0)			       (0x00000404 + 0x8*(i0))
#define NV01_RECT_RECTANGLE_SIZE__ESIZE				0x00000008
#define NV01_RECT_RECTANGLE_SIZE__LEN				0x00000010
#define NV01_RECT_RECTANGLE_SIZE_W__MASK			0x0000ffff
#define NV01_RECT_RECTANGLE_SIZE_W__SHIFT			0
#define NV01_RECT_RECTANGLE_SIZE_H__MASK			0xffff0000
#define NV01_RECT_RECTANGLE_SIZE_H__SHIFT			16


#define NV01_BLIT_PATCH						0x0000010c


#define NV15_BLIT_WAIT_FOR_IDLE					0x00000108

#define NV15_BLIT_FLIP_SET_READ					0x00000120

#define NV15_BLIT_FLIP_SET_WRITE				0x00000124

#define NV15_BLIT_FLIP_MAX					0x00000128

#define NV15_BLIT_FLIP_INCR_WRITE				0x0000012c

#define NV15_BLIT_FLIP_WAIT					0x00000130

#define NV15_BLIT_FLIP_CRTC_INCR_READ				0x00000134

#define NV01_BLIT_DMA_NOTIFY					0x00000180

#define NV01_BLIT_COLOR_KEY					0x00000184

#define NV04_BLIT_COLOR_KEY					0x00000184

#define NV01_BLIT_CLIP						0x00000188

#define NV01_BLIT_PATTERN					0x0000018c

#define NV04_BLIT_PATTERN					0x0000018c

#define NV01_BLIT_ROP						0x00000190

#define NV01_BLIT_BETA						0x00000194


#define NV01_BLIT_SURFACE_SRC					0x00000198

#define NV01_BLIT_SURFACE_DST					0x0000019c


#define NV04_BLIT_BETA4						0x00000198

#define NV04_BLIT_SURFACES					0x0000019c

#define NV01_BLIT_OPERATION					0x000002fc
#define NV01_BLIT_OPERATION_SRCCOPY_AND				0x00000000
#define NV01_BLIT_OPERATION_ROP_AND				0x00000001
#define NV01_BLIT_OPERATION_BLEND_AND				0x00000002
#define NV01_BLIT_OPERATION_SRCCOPY				0x00000003
#define NV01_BLIT_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_BLIT_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_BLIT_POINT_IN					0x00000300
#define NV01_BLIT_POINT_IN_X__MASK				0x0000ffff
#define NV01_BLIT_POINT_IN_X__SHIFT				0
#define NV01_BLIT_POINT_IN_Y__MASK				0xffff0000
#define NV01_BLIT_POINT_IN_Y__SHIFT				16

#define NV01_BLIT_POINT_OUT					0x00000304
#define NV01_BLIT_POINT_OUT_X__MASK				0x0000ffff
#define NV01_BLIT_POINT_OUT_X__SHIFT				0
#define NV01_BLIT_POINT_OUT_Y__MASK				0xffff0000
#define NV01_BLIT_POINT_OUT_Y__SHIFT				16

#define NV01_BLIT_SIZE						0x00000308
#define NV01_BLIT_SIZE_W__MASK					0x0000ffff
#define NV01_BLIT_SIZE_W__SHIFT					0
#define NV01_BLIT_SIZE_H__MASK					0xffff0000
#define NV01_BLIT_SIZE_H__SHIFT					16


#define NV04_INDEX_PATCH					0x0000010c

#define NV04_INDEX_DMA_NOTIFY					0x00000180

#define NV04_INDEX_DMA_LUT					0x00000184

#define NV04_INDEX_COLOR_KEY					0x00000188

#define NV04_INDEX_CLIP						0x0000018c

#define NV04_INDEX_PATTERN					0x00000190

#define NV04_INDEX_ROP						0x00000194

#define NV04_INDEX_BETA						0x00000198

#define NV04_INDEX_BETA4					0x0000019c

#define NV04_INDEX_SURFACE					0x000001a0

#define NV05_INDEX_SURFACE					0x000001a0

#define NV05_INDEX_COLOR_CONVERSION				0x000003e0

#define NV04_INDEX_OPERATION					0x000003e4

#define NV04_INDEX_COLOR_FORMAT					0x000003e8

#define NV04_INDEX_INDEX_FORMAT					0x000003ec

#define NV04_INDEX_LUT_OFFSET					0x000003f0

#define NV04_INDEX_POINT					0x000003f4

#define NV04_INDEX_SIZE_OUT					0x000003f8

#define NV04_INDEX_SIZE_IN					0x000003fc

#define NV04_INDEX_COLOR(i0)				       (0x00000400 + 0x4*(i0))
#define NV04_INDEX_COLOR__ESIZE					0x00000004
#define NV04_INDEX_COLOR__LEN					0x00000700


#define NV10_IFC_WAIT_FOR_IDLE					0x00000108

#define NV01_IFC_PATCH						0x0000010c

#define NV01_IFC_DMA_NOTIFY					0x00000180

#define NV01_IFC_COLOR_KEY					0x00000184

#define NV04_IFC_COLOR_KEY					0x00000184

#define NV01_IFC_CLIP						0x00000188

#define NV01_IFC_PATTERN					0x0000018c

#define NV04_IFC_PATTERN					0x0000018c

#define NV01_IFC_ROP						0x00000190

#define NV01_IFC_BETA						0x00000194


#define NV01_IFC_SURFACE_DST					0x00000198


#define NV04_IFC_BETA4						0x00000198

#define NV04_IFC_SURFACE					0x0000019c

#define NV05_IFC_COLOR_CONVERSION				0x000002f8

#define NV01_IFC_OPERATION					0x000002fc
#define NV01_IFC_OPERATION_SRCCOPY_AND				0x00000000
#define NV01_IFC_OPERATION_ROP_AND				0x00000001
#define NV01_IFC_OPERATION_BLEND_AND				0x00000002
#define NV01_IFC_OPERATION_SRCCOPY				0x00000003
#define NV01_IFC_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV01_IFC_OPERATION_BLEND_PREMULT			0x00000005

#define NV01_IFC_COLOR_FORMAT					0x00000300
#define NV01_IFC_COLOR_FORMAT_R5G6G5				0x00000001
#define NV01_IFC_COLOR_FORMAT_A1R5G5B5				0x00000002
#define NV01_IFC_COLOR_FORMAT_X1R5G5B5				0x00000003
#define NV01_IFC_COLOR_FORMAT_A8R8G8B8				0x00000004
#define NV01_IFC_COLOR_FORMAT_X8R8G8B8				0x00000005

#define NV01_IFC_POINT						0x00000304
#define NV01_IFC_POINT_X__MASK					0x0000ffff
#define NV01_IFC_POINT_X__SHIFT					0
#define NV01_IFC_POINT_Y__MASK					0xffff0000
#define NV01_IFC_POINT_Y__SHIFT					16

#define NV01_IFC_SIZE_OUT					0x00000308
#define NV01_IFC_SIZE_OUT_W__MASK				0x0000ffff
#define NV01_IFC_SIZE_OUT_W__SHIFT				0
#define NV01_IFC_SIZE_OUT_H__MASK				0xffff0000
#define NV01_IFC_SIZE_OUT_H__SHIFT				16

#define NV01_IFC_SIZE_IN					0x0000030c
#define NV01_IFC_SIZE_IN_W__MASK				0x0000ffff
#define NV01_IFC_SIZE_IN_W__SHIFT				0
#define NV01_IFC_SIZE_IN_H__MASK				0xffff0000
#define NV01_IFC_SIZE_IN_H__SHIFT				16

#define NV01_IFC_COLOR(i0)				       (0x00000400 + 0x4*(i0))
#define NV01_IFC_COLOR__ESIZE					0x00000004
#define NV01_IFC_COLOR__LEN					0x00000020

#define NV04_IFC_COLOR(i0)				       (0x00000400 + 0x4*(i0))
#define NV04_IFC_COLOR__ESIZE					0x00000004
#define NV04_IFC_COLOR__LEN					0x00000700


#define NV03_SIFC_PATCH						0x0000010c

#define NV03_SIFC_DMA_NOTIFY					0x00000180

#define NV03_SIFC_COLOR_KEY					0x00000184

#define NV04_SIFC_COLOR_KEY					0x00000184

#define NV03_SIFC_PATTERN					0x00000188

#define NV04_SIFC_PATTERN					0x00000188

#define NV03_SIFC_ROP						0x0000018c

#define NV03_SIFC_BETA						0x00000190


#define NV03_SIFC_SURFACE_DST					0x00000194


#define NV04_SIFC_BETA4						0x00000194

#define NV04_SIFC_SURFACE					0x00000198

#define NV05_SIFC_COLOR_CONVERSION				0x000002f8

#define NV03_SIFC_OPERATION					0x000002fc

#define NV03_SIFC_COLOR_FORMAT					0x00000300

#define NV03_SIFC_SIZE_IN					0x00000304
#define NV03_SIFC_SIZE_IN_W__MASK				0x0000ffff
#define NV03_SIFC_SIZE_IN_W__SHIFT				0
#define NV03_SIFC_SIZE_IN_H__MASK				0xffff0000
#define NV03_SIFC_SIZE_IN_H__SHIFT				16

#define NV03_SIFC_DX_DU						0x00000308

#define NV03_SIFC_DY_DV						0x0000030c

#define NV03_SIFC_CLIP_POINT					0x00000310
#define NV03_SIFC_CLIP_POINT_X__MASK				0x0000ffff
#define NV03_SIFC_CLIP_POINT_X__SHIFT				0
#define NV03_SIFC_CLIP_POINT_Y__MASK				0xffff0000
#define NV03_SIFC_CLIP_POINT_Y__SHIFT				16

#define NV03_SIFC_CLIP_SIZE					0x00000314
#define NV03_SIFC_CLIP_SIZE_W__MASK				0x0000ffff
#define NV03_SIFC_CLIP_SIZE_W__SHIFT				0
#define NV03_SIFC_CLIP_SIZE_H__MASK				0xffff0000
#define NV03_SIFC_CLIP_SIZE_H__SHIFT				16

#define NV03_SIFC_POINT12D4					0x00000318
#define NV03_SIFC_POINT12D4_X__MASK				0x0000ffff
#define NV03_SIFC_POINT12D4_X__SHIFT				0
#define NV03_SIFC_POINT12D4_Y__MASK				0xffff0000
#define NV03_SIFC_POINT12D4_Y__SHIFT				16

#define NV03_SIFC_COLOR(i0)				       (0x00000400 + 0x4*(i0))
#define NV03_SIFC_COLOR__ESIZE					0x00000004
#define NV03_SIFC_COLOR__LEN					0x00000700


#define NV10_SIFM_WAIT_FOR_IDLE					0x00000108

#define NV03_SIFM_DMA_NOTIFY					0x00000180

#define NV03_SIFM_DMA_IMAGE					0x00000184

#define NV03_SIFM_PATTERN					0x00000188

#define NV04_SIFM_PATTERN					0x00000188

#define NV03_SIFM_ROP						0x0000018c

#define NV03_SIFM_BETA						0x00000190


#define NV03_SIFM_SURFACE_DST					0x00000194


#define NV04_SIFM_BETA4						0x00000194

#define NV04_SIFM_SURFACE					0x00000198

#define NV05_SIFM_SURFACE					0x00000198

#define NV05_SIFM_COLOR_CONVERSION				0x000002fc
#define NV05_SIFM_COLOR_CONVERSION_DITHER			0x00000000
#define NV05_SIFM_COLOR_CONVERSION_TRUNCATE			0x00000001
#define NV05_SIFM_COLOR_CONVERSION_SUBTR_TRUNCATE		0x00000002

#define NV03_SIFM_COLOR_FORMAT					0x00000300
#define NV03_SIFM_COLOR_FORMAT_A1R5G5B5				0x00000001
#define NV03_SIFM_COLOR_FORMAT_X1R5G5B5				0x00000002
#define NV03_SIFM_COLOR_FORMAT_A8R8G8B8				0x00000003
#define NV03_SIFM_COLOR_FORMAT_X8R8G8B8				0x00000004
#define NV03_SIFM_COLOR_FORMAT_V8YB8U8YA8			0x00000005
#define NV03_SIFM_COLOR_FORMAT_YB8V8YA8U8			0x00000006
#define NV03_SIFM_COLOR_FORMAT_R5G6B5				0x00000007
#define NV03_SIFM_COLOR_FORMAT_Y8				0x00000008
#define NV03_SIFM_COLOR_FORMAT_AY8				0x00000009

#define NV03_SIFM_OPERATION					0x00000304
#define NV03_SIFM_OPERATION_SRCCOPY_AND				0x00000000
#define NV03_SIFM_OPERATION_ROP_AND				0x00000001
#define NV03_SIFM_OPERATION_BLEND_AND				0x00000002
#define NV03_SIFM_OPERATION_SRCCOPY				0x00000003
#define NV03_SIFM_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV03_SIFM_OPERATION_BLEND_PREMULT			0x00000005

#define NV03_SIFM_CLIP_POINT					0x00000308
#define NV03_SIFM_CLIP_POINT_X__MASK				0x0000ffff
#define NV03_SIFM_CLIP_POINT_X__SHIFT				0
#define NV03_SIFM_CLIP_POINT_Y__MASK				0xffff0000
#define NV03_SIFM_CLIP_POINT_Y__SHIFT				16

#define NV03_SIFM_CLIP_SIZE					0x0000030c
#define NV03_SIFM_CLIP_SIZE_W__MASK				0x0000ffff
#define NV03_SIFM_CLIP_SIZE_W__SHIFT				0
#define NV03_SIFM_CLIP_SIZE_H__MASK				0xffff0000
#define NV03_SIFM_CLIP_SIZE_H__SHIFT				16

#define NV03_SIFM_OUT_POINT					0x00000310
#define NV03_SIFM_OUT_POINT_X__MASK				0x0000ffff
#define NV03_SIFM_OUT_POINT_X__SHIFT				0
#define NV03_SIFM_OUT_POINT_Y__MASK				0xffff0000
#define NV03_SIFM_OUT_POINT_Y__SHIFT				16

#define NV03_SIFM_OUT_SIZE					0x00000314
#define NV03_SIFM_OUT_SIZE_W__MASK				0x0000ffff
#define NV03_SIFM_OUT_SIZE_W__SHIFT				0
#define NV03_SIFM_OUT_SIZE_H__MASK				0xffff0000
#define NV03_SIFM_OUT_SIZE_H__SHIFT				16

#define NV03_SIFM_DU_DX						0x00000318

#define NV03_SIFM_DV_DY						0x0000031c

#define NV03_SIFM_SIZE						0x00000400
#define NV03_SIFM_SIZE_W__MASK					0x0000ffff
#define NV03_SIFM_SIZE_W__SHIFT					0
#define NV03_SIFM_SIZE_H__MASK					0xffff0000
#define NV03_SIFM_SIZE_H__SHIFT					16

#define NV03_SIFM_FORMAT					0x00000404
#define NV03_SIFM_FORMAT_PITCH__MASK				0x0000ffff
#define NV03_SIFM_FORMAT_PITCH__SHIFT				0
#define NV03_SIFM_FORMAT_ORIGIN__MASK				0x00ff0000
#define NV03_SIFM_FORMAT_ORIGIN__SHIFT				16
#define NV03_SIFM_FORMAT_ORIGIN_CENTER				0x00010000
#define NV03_SIFM_FORMAT_ORIGIN_CORNER				0x00020000
#define NV03_SIFM_FORMAT_FILTER__MASK				0xff000000
#define NV03_SIFM_FORMAT_FILTER__SHIFT				24
#define NV03_SIFM_FORMAT_FILTER_POINT_SAMPLE			0x00000000
#define NV03_SIFM_FORMAT_FILTER_BILINEAR			0x01000000

#define NV03_SIFM_OFFSET					0x00000408

#define NV03_SIFM_POINT						0x0000040c
#define NV03_SIFM_POINT_U__MASK					0x0000ffff
#define NV03_SIFM_POINT_U__SHIFT				0
#define NV03_SIFM_POINT_V__MASK					0xffff0000
#define NV03_SIFM_POINT_V__SHIFT				16


#define NV50_SIFM_OFFSET_HIGH					0x00000410

#define NV50_SIFM_SRC_LINEAR					0x00000414

#define NV50_SIFM_SRC_TILE_MODE					0x00000418


#define NV03_GDI_DMA_NOTIFY					0x00000180

#define NV03_GDI_PATTERN					0x00000184

#define NV03_GDI_ROP						0x00000188

#define NV03_GDI_BETA						0x0000019c

#define NV03_GDI_SURFACE_DST					0x00000190

#define NV03_GDI_OPERATION					0x000002fc

#define NV03_GDI_COLOR_FORMAT					0x00000300

#define NV03_GDI_MONOCHROME_FORMAT				0x00000304

#define NV03_GDI_COLOR1_A					0x000003fc

#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT(i0)		       (0x00000400 + 0x8*(i0))
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT__ESIZE		0x00000008
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT__LEN			0x00000040
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT_Y__MASK		0x0000ffff
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT_Y__SHIFT		0
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT_X__MASK		0xffff0000
#define NV03_GDI_UNCLIPPED_RECTANGLE_POINT_X__SHIFT		16

#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE(i0)		       (0x00000404 + 0x8*(i0))
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE__ESIZE		0x00000008
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE__LEN			0x00000040
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE_H__MASK		0x0000ffff
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE_H__SHIFT		0
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE_W__MASK		0xffff0000
#define NV03_GDI_UNCLIPPED_RECTANGLE_SIZE_W__SHIFT		16

#define NV03_GDI_CLIP_POINT0_B					0x000007f4
#define NV03_GDI_CLIP_POINT0_B_L__MASK				0x0000ffff
#define NV03_GDI_CLIP_POINT0_B_L__SHIFT				0
#define NV03_GDI_CLIP_POINT0_B_T__MASK				0xffff0000
#define NV03_GDI_CLIP_POINT0_B_T__SHIFT				16

#define NV03_GDI_CLIP_POINT1_B					0x000007f8
#define NV03_GDI_CLIP_POINT1_B_R__MASK				0x0000ffff
#define NV03_GDI_CLIP_POINT1_B_R__SHIFT				0
#define NV03_GDI_CLIP_POINT1_B_B__MASK				0xffff0000
#define NV03_GDI_CLIP_POINT1_B_B__SHIFT				16

#define NV03_GDI_COLOR1_B					0x000007fc

#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0(i0)		       (0x00000800 + 0x8*(i0))
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0__ESIZE		0x00000008
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0__LEN			0x00000040
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0_L__MASK		0x0000ffff
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0_L__SHIFT		0
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0_T__MASK		0xffff0000
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_0_T__SHIFT		16

#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1(i0)		       (0x00000804 + 0x8*(i0))
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1__ESIZE		0x00000008
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1__LEN			0x00000040
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1_R__MASK		0x0000ffff
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1_R__SHIFT		0
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1_B__MASK		0xffff0000
#define NV03_GDI_CLIPPED_RECTANGLE_POINT_1_B__SHIFT		16

#define NV03_GDI_CLIP_C_POINT0					0x00000bec
#define NV03_GDI_CLIP_C_POINT0_L__MASK				0x0000ffff
#define NV03_GDI_CLIP_C_POINT0_L__SHIFT				0
#define NV03_GDI_CLIP_C_POINT0_T__MASK				0xffff0000
#define NV03_GDI_CLIP_C_POINT0_T__SHIFT				16

#define NV03_GDI_CLIP_C_POINT1					0x00000bf0
#define NV03_GDI_CLIP_C_POINT1_R__MASK				0x0000ffff
#define NV03_GDI_CLIP_C_POINT1_R__SHIFT				0
#define NV03_GDI_CLIP_C_POINT1_B__MASK				0xffff0000
#define NV03_GDI_CLIP_C_POINT1_B__SHIFT				16

#define NV03_GDI_COLOR1_C					0x00000bf4

#define NV03_GDI_SIZE_C						0x00000bf8
#define NV03_GDI_SIZE_C_W__MASK					0x0000ffff
#define NV03_GDI_SIZE_C_W__SHIFT				0
#define NV03_GDI_SIZE_C_H__MASK					0xffff0000
#define NV03_GDI_SIZE_C_H__SHIFT				16

#define NV03_GDI_POINT_C					0x00000bfc
#define NV03_GDI_POINT_C_X__MASK				0x0000ffff
#define NV03_GDI_POINT_C_X__SHIFT				0
#define NV03_GDI_POINT_C_Y__MASK				0xffff0000
#define NV03_GDI_POINT_C_Y__SHIFT				16

#define NV03_GDI_MONOCHROME_COLOR1_C(i0)		       (0x00000c00 + 0x4*(i0))
#define NV03_GDI_MONOCHROME_COLOR1_C__ESIZE			0x00000004
#define NV03_GDI_MONOCHROME_COLOR1_C__LEN			0x00000080

#define NV03_GDI_CLIP_D_POINT0					0x00000fe8
#define NV03_GDI_CLIP_D_POINT0_L__MASK				0x0000ffff
#define NV03_GDI_CLIP_D_POINT0_L__SHIFT				0
#define NV03_GDI_CLIP_D_POINT0_T__MASK				0xffff0000
#define NV03_GDI_CLIP_D_POINT0_T__SHIFT				16

#define NV03_GDI_CLIP_D_POINT1					0x00000fec
#define NV03_GDI_CLIP_D_POINT1_R__MASK				0x0000ffff
#define NV03_GDI_CLIP_D_POINT1_R__SHIFT				0
#define NV03_GDI_CLIP_D_POINT1_B__MASK				0xffff0000
#define NV03_GDI_CLIP_D_POINT1_B__SHIFT				16

#define NV03_GDI_COLOR1_D					0x00000ff0

#define NV03_GDI_SIZE_IN_D					0x00000ff4
#define NV03_GDI_SIZE_IN_D_W__MASK				0x0000ffff
#define NV03_GDI_SIZE_IN_D_W__SHIFT				0
#define NV03_GDI_SIZE_IN_D_H__MASK				0xffff0000
#define NV03_GDI_SIZE_IN_D_H__SHIFT				16

#define NV03_GDI_SIZE_OUT_D					0x00000ff8
#define NV03_GDI_SIZE_OUT_D_W__MASK				0x0000ffff
#define NV03_GDI_SIZE_OUT_D_W__SHIFT				0
#define NV03_GDI_SIZE_OUT_D_H__MASK				0xffff0000
#define NV03_GDI_SIZE_OUT_D_H__SHIFT				16

#define NV03_GDI_POINT_D					0x00000ffc
#define NV03_GDI_POINT_D_X__MASK				0x0000ffff
#define NV03_GDI_POINT_D_X__SHIFT				0
#define NV03_GDI_POINT_D_Y__MASK				0xffff0000
#define NV03_GDI_POINT_D_Y__SHIFT				16

#define NV03_GDI_MONOCHROME_COLOR1_D(i0)		       (0x00001000 + 0x4*(i0))
#define NV03_GDI_MONOCHROME_COLOR1_D__ESIZE			0x00000004
#define NV03_GDI_MONOCHROME_COLOR1_D__LEN			0x00000080

#define NV03_GDI_CLIP_E_POINT0					0x000013e4
#define NV03_GDI_CLIP_E_POINT0_L__MASK				0x0000ffff
#define NV03_GDI_CLIP_E_POINT0_L__SHIFT				0
#define NV03_GDI_CLIP_E_POINT0_T__MASK				0xffff0000
#define NV03_GDI_CLIP_E_POINT0_T__SHIFT				16

#define NV03_GDI_CLIP_E_POINT1					0x000013e8
#define NV03_GDI_CLIP_E_POINT1_R__MASK				0x0000ffff
#define NV03_GDI_CLIP_E_POINT1_R__SHIFT				0
#define NV03_GDI_CLIP_E_POINT1_B__MASK				0xffff0000
#define NV03_GDI_CLIP_E_POINT1_B__SHIFT				16

#define NV03_GDI_COLOR0_E					0x000013ec

#define NV03_GDI_COLOR1_E					0x000013f0

#define NV03_GDI_SIZE_IN_E					0x000013f4
#define NV03_GDI_SIZE_IN_E_W__MASK				0x0000ffff
#define NV03_GDI_SIZE_IN_E_W__SHIFT				0
#define NV03_GDI_SIZE_IN_E_H__MASK				0xffff0000
#define NV03_GDI_SIZE_IN_E_H__SHIFT				16

#define NV03_GDI_SIZE_OUT_E					0x000013f8
#define NV03_GDI_SIZE_OUT_E_W__MASK				0x0000ffff
#define NV03_GDI_SIZE_OUT_E_W__SHIFT				0
#define NV03_GDI_SIZE_OUT_E_H__MASK				0xffff0000
#define NV03_GDI_SIZE_OUT_E_H__SHIFT				16

#define NV03_GDI_POINT_E					0x000013fc
#define NV03_GDI_POINT_E_X__MASK				0x0000ffff
#define NV03_GDI_POINT_E_X__SHIFT				0
#define NV03_GDI_POINT_E_Y__MASK				0xffff0000
#define NV03_GDI_POINT_E_Y__SHIFT				16

#define NV03_GDI_MONOCHROME_COLOR01_E(i0)		       (0x00001400 + 0x4*(i0))
#define NV03_GDI_MONOCHROME_COLOR01_E__ESIZE			0x00000004
#define NV03_GDI_MONOCHROME_COLOR01_E__LEN			0x00000080


#define NV04_GDI_PATCH						0x0000010c

#define NV04_GDI_DMA_NOTIFY					0x00000180

#define NV04_GDI_DMA_FONTS					0x00000184

#define NV04_GDI_PATTERN					0x00000188

#define NV04_GDI_ROP						0x0000018c

#define NV04_GDI_BETA						0x00000190

#define NV04_GDI_BETA4						0x00000194

#define NV04_GDI_SURFACE					0x00000198

#define NV04_GDI_OPERATION					0x000002fc
#define NV04_GDI_OPERATION_SRCCOPY_AND				0x00000000
#define NV04_GDI_OPERATION_ROP_AND				0x00000001
#define NV04_GDI_OPERATION_BLEND_AND				0x00000002
#define NV04_GDI_OPERATION_SRCCOPY				0x00000003
#define NV04_GDI_OPERATION_SRCCOPY_PREMULT			0x00000004
#define NV04_GDI_OPERATION_BLEND_PREMULT			0x00000005

#define NV04_GDI_COLOR_FORMAT					0x00000300
#define NV04_GDI_COLOR_FORMAT_A16R5G6B5				0x00000001
#define NV04_GDI_COLOR_FORMAT_X16A1R5G5B5			0x00000002
#define NV04_GDI_COLOR_FORMAT_A8R8G8B8				0x00000003

#define NV04_GDI_MONOCHROME_FORMAT				0x00000304
#define NV04_GDI_MONOCHROME_FORMAT_CGA6				0x00000001
#define NV04_GDI_MONOCHROME_FORMAT_LE				0x00000002

#define NV04_GDI_COLOR1_A					0x000003fc

#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT(i0)		       (0x00000400 + 0x8*(i0))
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT__ESIZE		0x00000008
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT__LEN			0x00000020
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT_Y__MASK		0x0000ffff
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT_Y__SHIFT		0
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT_X__MASK		0xffff0000
#define NV04_GDI_UNCLIPPED_RECTANGLE_POINT_X__SHIFT		16

#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE(i0)		       (0x00000404 + 0x8*(i0))
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE__ESIZE		0x00000008
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE__LEN			0x00000020
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE_H__MASK		0x0000ffff
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE_H__SHIFT		0
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE_W__MASK		0xffff0000
#define NV04_GDI_UNCLIPPED_RECTANGLE_SIZE_W__SHIFT		16

#define NV04_GDI_CLIP_B_POINT0					0x000005f4
#define NV04_GDI_CLIP_B_POINT0_L__MASK				0x0000ffff
#define NV04_GDI_CLIP_B_POINT0_L__SHIFT				0
#define NV04_GDI_CLIP_B_POINT0_T__MASK				0xffff0000
#define NV04_GDI_CLIP_B_POINT0_T__SHIFT				16

#define NV04_GDI_CLIP_B_POINT1					0x000005f8
#define NV04_GDI_CLIP_B_POINT1_R__MASK				0x0000ffff
#define NV04_GDI_CLIP_B_POINT1_R__SHIFT				0
#define NV04_GDI_CLIP_B_POINT1_B__MASK				0xffff0000
#define NV04_GDI_CLIP_B_POINT1_B__SHIFT				16

#define NV04_GDI_COLOR1_B					0x000005fc

#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0(i0)		       (0x00000600 + 0x8*(i0))
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0__ESIZE		0x00000008
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0__LEN			0x00000020
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0_L__MASK		0x0000ffff
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0_L__SHIFT		0
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0_T__MASK		0xffff0000
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_0_T__SHIFT		16

#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1(i0)		       (0x00000604 + 0x8*(i0))
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1__ESIZE		0x00000008
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1__LEN			0x00000020
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1_R__MASK		0x0000ffff
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1_R__SHIFT		0
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1_B__MASK		0xffff0000
#define NV04_GDI_CLIPPED_RECTANGLE_POINT_1_B__SHIFT		16

#define NV04_GDI_CLIP_C_POINT0					0x000007ec
#define NV04_GDI_CLIP_C_POINT0_L__MASK				0x0000ffff
#define NV04_GDI_CLIP_C_POINT0_L__SHIFT				0
#define NV04_GDI_CLIP_C_POINT0_T__MASK				0xffff0000
#define NV04_GDI_CLIP_C_POINT0_T__SHIFT				16

#define NV04_GDI_CLIP_C_POINT1					0x000007f0
#define NV04_GDI_CLIP_C_POINT1_R__MASK				0x0000ffff
#define NV04_GDI_CLIP_C_POINT1_R__SHIFT				0
#define NV04_GDI_CLIP_C_POINT1_B__MASK				0xffff0000
#define NV04_GDI_CLIP_C_POINT1_B__SHIFT				16

#define NV04_GDI_COLOR1_C					0x000007f4

#define NV04_GDI_SIZE_C						0x000007f8
#define NV04_GDI_SIZE_C_W__MASK					0x0000ffff
#define NV04_GDI_SIZE_C_W__SHIFT				0
#define NV04_GDI_SIZE_C_H__MASK					0xffff0000
#define NV04_GDI_SIZE_C_H__SHIFT				16

#define NV04_GDI_POINT_C					0x000007fc
#define NV04_GDI_POINT_C_X__MASK				0x0000ffff
#define NV04_GDI_POINT_C_X__SHIFT				0
#define NV04_GDI_POINT_C_Y__MASK				0xffff0000
#define NV04_GDI_POINT_C_Y__SHIFT				16

#define NV04_GDI_MONOCHROME_COLOR1_C(i0)		       (0x00000800 + 0x4*(i0))
#define NV04_GDI_MONOCHROME_COLOR1_C__ESIZE			0x00000004
#define NV04_GDI_MONOCHROME_COLOR1_C__LEN			0x00000080

#define NV04_GDI_CLIP_E_POINT0					0x00000be4
#define NV04_GDI_CLIP_E_POINT0_L__MASK				0x0000ffff
#define NV04_GDI_CLIP_E_POINT0_L__SHIFT				0
#define NV04_GDI_CLIP_E_POINT0_T__MASK				0xffff0000
#define NV04_GDI_CLIP_E_POINT0_T__SHIFT				16

#define NV04_GDI_CLIP_E_POINT1					0x00000be8
#define NV04_GDI_CLIP_E_POINT1_R__MASK				0x0000ffff
#define NV04_GDI_CLIP_E_POINT1_R__SHIFT				0
#define NV04_GDI_CLIP_E_POINT1_B__MASK				0xffff0000
#define NV04_GDI_CLIP_E_POINT1_B__SHIFT				16

#define NV04_GDI_COLOR0_E					0x00000bec

#define NV04_GDI_COLOR1_E					0x00000bf0

#define NV04_GDI_SIZE_IN_E					0x00000bf4
#define NV04_GDI_SIZE_IN_E_W__MASK				0x0000ffff
#define NV04_GDI_SIZE_IN_E_W__SHIFT				0
#define NV04_GDI_SIZE_IN_E_H__MASK				0xffff0000
#define NV04_GDI_SIZE_IN_E_H__SHIFT				16

#define NV04_GDI_SIZE_OUT_E					0x00000bf8
#define NV04_GDI_SIZE_OUT_E_W__MASK				0x0000ffff
#define NV04_GDI_SIZE_OUT_E_W__SHIFT				0
#define NV04_GDI_SIZE_OUT_E_H__MASK				0xffff0000
#define NV04_GDI_SIZE_OUT_E_H__SHIFT				16

#define NV04_GDI_POINT_E					0x00000bfc
#define NV04_GDI_POINT_E_X__MASK				0x0000ffff
#define NV04_GDI_POINT_E_X__SHIFT				0
#define NV04_GDI_POINT_E_Y__MASK				0xffff0000
#define NV04_GDI_POINT_E_Y__SHIFT				16

#define NV04_GDI_MONOCHROME_COLOR01_E(i0)		       (0x00000c00 + 0x4*(i0))
#define NV04_GDI_MONOCHROME_COLOR01_E__ESIZE			0x00000004
#define NV04_GDI_MONOCHROME_COLOR01_E__LEN			0x00000080

#define NV04_GDI_FONT_F						0x00000ff0
#define NV04_GDI_FONT_F_OFFSET__MASK				0x0fffffff
#define NV04_GDI_FONT_F_OFFSET__SHIFT				0
#define NV04_GDI_FONT_F_PITCH__MASK				0xf0000000
#define NV04_GDI_FONT_F_PITCH__SHIFT				28

#define NV04_GDI_CLIP_F_POINT0					0x00000ff4
#define NV04_GDI_CLIP_F_POINT0_L__MASK				0x0000ffff
#define NV04_GDI_CLIP_F_POINT0_L__SHIFT				0
#define NV04_GDI_CLIP_F_POINT0_T__MASK				0xffff0000
#define NV04_GDI_CLIP_F_POINT0_T__SHIFT				16

#define NV04_GDI_CLIP_F_POINT1					0x00000ff8
#define NV04_GDI_CLIP_F_POINT1_R__MASK				0x0000ffff
#define NV04_GDI_CLIP_F_POINT1_R__SHIFT				0
#define NV04_GDI_CLIP_F_POINT1_B__MASK				0xffff0000
#define NV04_GDI_CLIP_F_POINT1_B__SHIFT				16

#define NV04_GDI_COLOR1_F					0x00000ffc

#define NV04_GDI_CHARACTER_COLOR1_F(i0)			       (0x00001000 + 0x4*(i0))
#define NV04_GDI_CHARACTER_COLOR1_F__ESIZE			0x00000004
#define NV04_GDI_CHARACTER_COLOR1_F__LEN			0x00000100
#define NV04_GDI_CHARACTER_COLOR1_F_INDEX__MASK			0x000000ff
#define NV04_GDI_CHARACTER_COLOR1_F_INDEX__SHIFT		0
#define NV04_GDI_CHARACTER_COLOR1_F_X__MASK			0x000fff00
#define NV04_GDI_CHARACTER_COLOR1_F_X__SHIFT			8
#define NV04_GDI_CHARACTER_COLOR1_F_Y__MASK			0xfff00000
#define NV04_GDI_CHARACTER_COLOR1_F_Y__SHIFT			20

#define NV04_GDI_FONT_G						0x000017f0
#define NV04_GDI_FONT_G_OFFSET__MASK				0x0fffffff
#define NV04_GDI_FONT_G_OFFSET__SHIFT				0
#define NV04_GDI_FONT_G_PITCH__MASK				0xf0000000
#define NV04_GDI_FONT_G_PITCH__SHIFT				28

#define NV04_GDI_CLIP_G_POINT0					0x000017f4
#define NV04_GDI_CLIP_G_POINT0_L__MASK				0x0000ffff
#define NV04_GDI_CLIP_G_POINT0_L__SHIFT				0
#define NV04_GDI_CLIP_G_POINT0_T__MASK				0xffff0000
#define NV04_GDI_CLIP_G_POINT0_T__SHIFT				16

#define NV04_GDI_CLIP_G_POINT1					0x000017f8
#define NV04_GDI_CLIP_G_POINT1_R__MASK				0x0000ffff
#define NV04_GDI_CLIP_G_POINT1_R__SHIFT				0
#define NV04_GDI_CLIP_G_POINT1_B__MASK				0xffff0000
#define NV04_GDI_CLIP_G_POINT1_B__SHIFT				16

#define NV04_GDI_COLOR1_G					0x000017fc

#define NV04_GDI_CHARACTER_COLOR1_G_POINT(i0)		       (0x00001800 + 0x8*(i0))
#define NV04_GDI_CHARACTER_COLOR1_G_POINT__ESIZE		0x00000008
#define NV04_GDI_CHARACTER_COLOR1_G_POINT__LEN			0x00000100
#define NV04_GDI_CHARACTER_COLOR1_G_POINT_X__MASK		0x0000ffff
#define NV04_GDI_CHARACTER_COLOR1_G_POINT_X__SHIFT		0
#define NV04_GDI_CHARACTER_COLOR1_G_POINT_Y__MASK		0xffff0000
#define NV04_GDI_CHARACTER_COLOR1_G_POINT_Y__SHIFT		16

#define NV04_GDI_CHARACTER_COLOR1_G_INDEX(i0)		       (0x00001804 + 0x8*(i0))
#define NV04_GDI_CHARACTER_COLOR1_G_INDEX__ESIZE		0x00000008
#define NV04_GDI_CHARACTER_COLOR1_G_INDEX__LEN			0x00000100


#define NV10_TEXUPLOAD_WAIT_FOR_IDLE				0x00000108

#define NV10_TEXUPLOAD_DMA_NOTIFY				0x00000180

#define NV10_TEXUPLOAD_SURFACE					0x00000184

#define NV10_TEXUPLOAD_COLOR_FORMAT				0x00000300

#define NV10_TEXUPLOAD_POINT					0x00000304
#define NV10_TEXUPLOAD_POINT_X__MASK				0x0000ffff
#define NV10_TEXUPLOAD_POINT_X__SHIFT				0
#define NV10_TEXUPLOAD_POINT_Y__MASK				0xffff0000
#define NV10_TEXUPLOAD_POINT_Y__SHIFT				16

#define NV10_TEXUPLOAD_SIZE					0x00000308
#define NV10_TEXUPLOAD_SIZE_W__MASK				0x0000ffff
#define NV10_TEXUPLOAD_SIZE_W__SHIFT				0
#define NV10_TEXUPLOAD_SIZE_H__MASK				0xffff0000
#define NV10_TEXUPLOAD_SIZE_H__SHIFT				16

#define NV10_TEXUPLOAD_CLIP_HORIZONTAL				0x0000030c
#define NV10_TEXUPLOAD_CLIP_HORIZONTAL_X__MASK			0x0000ffff
#define NV10_TEXUPLOAD_CLIP_HORIZONTAL_X__SHIFT			0
#define NV10_TEXUPLOAD_CLIP_HORIZONTAL_W__MASK			0xffff0000
#define NV10_TEXUPLOAD_CLIP_HORIZONTAL_W__SHIFT			16

#define NV10_TEXUPLOAD_CLIP_VERTICAL				0x00000310
#define NV10_TEXUPLOAD_CLIP_VERTICAL_Y__MASK			0x0000ffff
#define NV10_TEXUPLOAD_CLIP_VERTICAL_Y__SHIFT			0
#define NV10_TEXUPLOAD_CLIP_VERTICAL_H__MASK			0xffff0000
#define NV10_TEXUPLOAD_CLIP_VERTICAL_H__SHIFT			16

#define NV10_TEXUPLOAD_COLOR(i0)			       (0x00000400 + 0x4*(i0))
#define NV10_TEXUPLOAD_COLOR__ESIZE				0x00000004
#define NV10_TEXUPLOAD_COLOR__LEN				0x00000700


#endif /* _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NV01_2D_XML */
