dp3(8)          g5<1>.xF        g5<4>.xyzzF     g5<4>.xyzzF     { align16 };
dp3(8)          m5<1>.xF        g3<0>.xyzzF     g6<4>.xyzzF     { align16 NoDDClr };
dp3(8)          m5<1>.yF        g3.4<0>.xyzzF   g6<4>.xyzzF     { align16 NoDDClr,NoDDChk };
dp3(8)          g19<1>.xF       g3<0>.xyzzF     g3.4<0>.xyzzF   { align16 NoDDClr };
dp3(8)          g19<1>.yF       g3<0>.xyzzF     g4<0>.xyzzF     { align16 NoDDClr,NoDDChk };
dp3(8)          g19<1>.zF       g3<0>.xyzzF     g4.4<0>.xyzzF   { align16 NoDDChk };
dp3(8)          m5<1>.xF        g4<4>.xyzzF     g5<4>.xyzzF     { align16 };
dp3.le.f0.0(8)  g18<1>.xF       g17<4>.xyzzF    g3.4<0>.xyzzF   { align16 };
dp3.sat(8)      g4<1>.xF        g4<4>.xyzzF     g5<4>.xyzzF     { align16 };
